#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.enums as enums
import octobot_commons.constants as constants
import octobot_commons.databases as databases
import octobot_commons.time_frame_manager as time_frame_manager
import octobot_tentacles_manager.api as tentacles_manager_api


def get_eval_time(full_candle=None, time_frame=None, partial_candle=None, kline=None):
    if full_candle is not None and time_frame is not None:
        # add one full time frame seconds since a full candle is available when the next has started
        return full_candle[enums.PriceIndexes.IND_PRICE_TIME.value] + \
               enums.TimeFramesMinutes[enums.TimeFrames(time_frame)] * constants.MINUTE_TO_SECONDS
    if partial_candle is not None:
        return partial_candle[enums.PriceIndexes.IND_PRICE_TIME.value]
    if kline is not None:
        return kline[enums.PriceIndexes.IND_PRICE_TIME.value]
    raise ValueError("Invalid arguments")


def get_shortest_time_frame(ideal_time_frame, preferred_available_time_frames, others):
    if ideal_time_frame in preferred_available_time_frames:
        return ideal_time_frame
    if preferred_available_time_frames:
        return time_frame_manager.sort_time_frames(preferred_available_time_frames)[0]
    else:
        return time_frame_manager.sort_time_frames(others)[0]


def local_trading_context(evaluator, symbol, time_frame, trigger_cache_timestamp,
                          cryptocurrency=None, exchange=None, exchange_id=None,
                          trigger_source=None, trigger_value=None):
    try:
        import octobot_trading.api as exchange_api
        import octobot_trading.modes as modes
        exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
            exchange or evaluator.exchange_name,
            exchange_id or exchange_api.get_exchange_id_from_matrix_id(evaluator.exchange_name, evaluator.matrix_id)
        )
        trading_modes = exchange_api.get_trading_modes(exchange_manager)
        return modes.Context(
            evaluator,
            exchange_manager,
            exchange_api.get_trader(exchange_manager),
            exchange or evaluator.exchange_name,
            symbol,
            evaluator.matrix_id,
            cryptocurrency,
            symbol,
            time_frame,
            evaluator.logger,
            trading_modes[0].__class__,
            trigger_cache_timestamp,
            trigger_source,
            trigger_value,
            None,
            None,
        )
    except ImportError:
        evaluator.logger.error("OctoBot-Evaluator local_trading_context requires OctoBot-Trading package installed")
        raise


def local_cache_client(evaluator, symbol, time_frame, exchange_name=None):
    try:
        exchange_name = exchange_name or evaluator.exchange_name
        import octobot_trading.api as exchange_api
        exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
            exchange_name,
            exchange_api.get_exchange_id_from_matrix_id(exchange_name, evaluator.matrix_id)
        )
        return databases.CacheClient(evaluator, exchange_name, symbol, time_frame,
                                     evaluator.tentacles_setup_config,
                                     not exchange_api.get_is_backtesting(exchange_manager))
    except ImportError:
        evaluator.logger.error("OctoBot-Evaluator local_cache_client requires OctoBot-Trading package installed")
        raise


def get_required_candles_count(trading_mode_class, tentacles_setup_config):
    return tentacles_manager_api.get_tentacle_config(tentacles_setup_config, trading_mode_class).get(
        constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT,
        constants.DEFAULT_IGNORED_VALUE
    )



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock

import octobot_tentacles_manager.api as tentacles_api
import octobot_evaluators.evaluators as evaluators
import octobot_commons.enums as enums
import octobot_commons.constants as constants
import octobot_commons.symbols.symbol_util as symbol_util
import octobot_evaluators.evaluators.evaluator_factory as evaluator_factory
from octobot_evaluators.evaluators.evaluator_factory import _extract_traded_pairs, _filter_pairs

from tests import event_loop, matrix_id, install_tentacles, evaluators_and_matrix_channels

pytestmark = pytest.mark.asyncio

exchange_name = "TEST_EXCHANGE_NAME"
bot_id = "TEST_BOT_ID"
symbols_by_crypto_currencies = {
    "Bitcoin": ["BTC/USDT"],
    "Ethereum": ["ETH/USD", "ETH/BTC"]
}
symbols = ["BTC/USDT", "ETH/USD", "ETH/BTC"]
time_frames = [enums.TimeFrames.ONE_HOUR, enums.TimeFrames.FOUR_HOURS]

crypto_currency_name_by_crypto_currencies = {}
symbols_by_crypto_currency_tickers = {}
for name, symbol_list in symbols_by_crypto_currencies.items():
    ticker = symbol_util.parse_symbol(symbol_list[0]).base
    crypto_currency_name_by_crypto_currencies[ticker] = name
    symbols_by_crypto_currency_tickers[ticker] = symbol_list


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_all_type_evaluators(evaluators_and_matrix_channels):
    tentacles_setup_config = tentacles_api.get_tentacles_setup_config()
    with mock.patch.object(evaluator_factory, "_start_evaluators", mock.AsyncMock()) as _start_evaluators_mock:
        created_evaluators = await evaluators.create_and_start_all_type_evaluators(tentacles_setup_config,
                                                                                   matrix_id=evaluators_and_matrix_channels,
                                                                                   exchange_name=exchange_name,
                                                                                   bot_id=bot_id,
                                                                                   symbols_by_crypto_currencies=symbols_by_crypto_currencies,
                                                                                   symbols=symbols,
                                                                                   time_frames=time_frames)
        _start_evaluators_mock.assert_not_called()

    assert not created_evaluators  # Trading package is not installed


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_strategy_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.StrategyEvaluator, evaluators_and_matrix_channels, [
        tentacles.SimpleStrategyEvaluator,
        tentacles.DipAnalyserStrategyEvaluator,
        tentacles.MoveSignalsStrategyEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_ta_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.TAEvaluator, evaluators_and_matrix_channels, [
        tentacles.RSIMomentumEvaluator,
        tentacles.ADXMomentumEvaluator,
        tentacles.StochasticRSIVolatilityEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_social_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.SocialEvaluator, evaluators_and_matrix_channels, [
        tentacles.RedditForumEvaluator
    ])


@pytest.mark.usefixtures("event_loop", "install_tentacles")
async def test_create_rt_evaluators(evaluators_and_matrix_channels):
    import tentacles
    await _test_evaluators_creation(evaluators.RealTimeEvaluator, evaluators_and_matrix_channels, [
        tentacles.InstantFluctuationsEvaluator
    ])


async def _test_evaluators_creation(evaluator_parent_class, fixture_matrix_id, expected_evaluators):
    tentacles_setup_config = tentacles_api.get_tentacles_setup_config()

    # activate all evaluators in tentacle config
    for tentacle_type_key, tentacle_type_value in tentacles_setup_config.tentacles_activation.items():
        for tentacle_name in tentacle_type_value:
            tentacles_setup_config.tentacles_activation[tentacle_type_key][tentacle_name] = True

    # mock start method to prevent side effects (octobot-trading imports, etc)
    created_evaluators = await evaluators.create_evaluators(evaluator_parent_class=evaluator_parent_class,
                                                            tentacles_setup_config=tentacles_setup_config,
                                                            matrix_id=fixture_matrix_id,
                                                            exchange_name=exchange_name,
                                                            bot_id=bot_id,
                                                            crypto_currency_name_by_crypto_currencies=crypto_currency_name_by_crypto_currencies,
                                                            symbols_by_crypto_currency_tickers=symbols_by_crypto_currency_tickers,
                                                            symbols=symbols,
                                                            time_frames=time_frames)
    assert created_evaluators
    assert all([evaluator.__class__ in expected_evaluators for evaluator in created_evaluators])


async def test_start_evaluators():
    eval_mock = mock.Mock()
    eval_mock.start_evaluator = mock.AsyncMock()
    with mock.patch.object(evaluator_factory, "_prioritized_evaluators", mock.Mock(return_value=[eval_mock])) \
        as prioritized_evaluators_mock:
        await evaluator_factory._start_evaluators([[1, 2], [3, None]], "tentacles_setup_config", "bot_id")
        prioritized_evaluators_mock.assert_called_once_with([1, 2, 3], "tentacles_setup_config")
        eval_mock.start_evaluator.assert_called_once_with("bot_id")


async def test_prioritized_evaluators():
    eval_mock_1 = mock.Mock()
    eval_mock_1.get_evaluator_priority = mock.Mock(return_value=constants.DEFAULT_EVALUATOR_PRIORITY)
    eval_mock_2 = mock.Mock()
    eval_mock_2.get_evaluator_priority = mock.Mock(return_value=constants.DEFAULT_EVALUATOR_PRIORITY)
    eval_mock_3 = mock.Mock()
    eval_mock_3.get_evaluator_priority = mock.Mock(return_value=constants.DEFAULT_EVALUATOR_PRIORITY)
    assert evaluator_factory._prioritized_evaluators(
        [eval_mock_1, eval_mock_2, eval_mock_3],
        "tentacles_setup_config") == [eval_mock_1, eval_mock_2, eval_mock_3]
    eval_mock_1.get_evaluator_priority = mock.Mock(return_value=1)
    eval_mock_2.get_evaluator_priority = mock.Mock(return_value=-5.6)
    assert evaluator_factory._prioritized_evaluators(
        [eval_mock_1, eval_mock_2, eval_mock_3],
        "tentacles_setup_config") == [eval_mock_1, eval_mock_3, eval_mock_2]
    eval_mock_2 = mock.Mock()
    eval_mock_2.get_evaluator_priority = mock.Mock(return_value=5.6)
    assert evaluator_factory._prioritized_evaluators(
        [eval_mock_1, eval_mock_2, eval_mock_3],
        "tentacles_setup_config") == [eval_mock_2, eval_mock_1, eval_mock_3]


async def test_extract_traded_pairs():
    exchange_name = "binance"
    matrix_id = "id"
    exchange_api = ExchangeAPIMock()

    # no symbol config
    symbols_by_crypto_currencies = None
    crypto_currency_name_by_crypto_currencies, symbols_by_crypto_currency_tickers = \
        _extract_traded_pairs(symbols_by_crypto_currencies, exchange_name, matrix_id, exchange_api)
    assert crypto_currency_name_by_crypto_currencies == {}
    assert symbols_by_crypto_currency_tickers == {}

    # normal symbol config
    symbols_by_crypto_currencies = {
        'AAVE': ['AAVE/BTC', 'AAVE/USDT'],
        'Cardano': ['ADA/BTC']
    }
    crypto_currency_name_by_crypto_currencies, symbols_by_crypto_currency_tickers = \
        _extract_traded_pairs(symbols_by_crypto_currencies, exchange_name, matrix_id, exchange_api)
    assert crypto_currency_name_by_crypto_currencies == {
        'AAVE': 'AAVE',
        'ADA': 'Cardano'
    }
    assert symbols_by_crypto_currency_tickers == {
        'AAVE': {'AAVE/BTC', 'AAVE/USDT'},
        'ADA': {'ADA/BTC'}
    }

    # AAVE/USDT in Cardano symbol config
    symbols_by_crypto_currencies = {
        'AAVE': ['AAVE/BTC'],
        'Cardano': ['AAVE/USDT', 'ADA/BTC']
    }
    crypto_currency_name_by_crypto_currencies, symbols_by_crypto_currency_tickers = \
        _extract_traded_pairs(symbols_by_crypto_currencies, exchange_name, matrix_id, exchange_api)
    assert crypto_currency_name_by_crypto_currencies == {
        'AAVE': 'AAVE',
        'ADA': 'Cardano'
    }
    assert symbols_by_crypto_currency_tickers == {
        'AAVE': {'AAVE/BTC', 'AAVE/USDT'},
        'ADA': {'ADA/BTC'}
    }

    # Many symbol config by reference market
    symbols_by_crypto_currencies = {
        'Bitcoin': [
            'AAVE/BTC', 'ADA/BTC', 'ATOM/BTC', 'BAT/BTC', 'BNB/BTC', 'DASH/BTC', 'DOT/BTC',
            'EOS/BTC', 'ETC/BTC', 'ETH/BTC', 'FIL/BTC', 'LINK/BTC', 'LTC/BTC', 'NEO/BTC',
            'ONT/BTC', 'ROSE/BTC', 'SUSHI/BTC', 'SXP/BTC', 'THETA/BTC', 'TOMO/BTC', 'UNI/BTC',
            'WAN/BTC', 'XLM/BTC', 'XMR/BTC', 'XTZ/BTC', 'YFI/BTC'
        ],
        'Tether': [
            'AAVE/USDT', 'ADA/USDT', 'ATOM/USDT', 'BAT/USDT', 'BNB/USDT', 'BTC/USDT',
            'DASH/USDT', 'DOT/USDT', 'EOS/USDT', 'ETC/USDT', 'ETH/USDT', 'FIL/USDT',
            'LINK/USDT', 'LTC/USDT', 'NEO/USDT', 'ONT/USDT', 'ROSE/USDT', 'SUSHI/USDT',
            'SXP/USDT', 'THETA/USDT', 'TOMO/USDT', 'UNI/USDT', 'WAN/USDT', 'XLM/USDT',
            'XMR/USDT', 'XTZ/USDT', 'YFI/USDT'
        ]
    }
    crypto_currency_name_by_crypto_currencies, symbols_by_crypto_currency_tickers = \
        _extract_traded_pairs(symbols_by_crypto_currencies, exchange_name, matrix_id, exchange_api)
    assert crypto_currency_name_by_crypto_currencies == {
        'AAVE': 'Bitcoin', 'ADA': 'Bitcoin', 'ATOM': 'Bitcoin',
        'BAT': 'Bitcoin', 'BNB': 'Bitcoin', 'DASH': 'Bitcoin',
        'DOT': 'Bitcoin', 'EOS': 'Bitcoin', 'ETC': 'Bitcoin',
        'ETH': 'Bitcoin', 'FIL': 'Bitcoin', 'LINK': 'Bitcoin',
        'LTC': 'Bitcoin', 'NEO': 'Bitcoin', 'ONT': 'Bitcoin',
        'ROSE': 'Bitcoin', 'SUSHI': 'Bitcoin', 'SXP': 'Bitcoin',
        'THETA': 'Bitcoin', 'TOMO': 'Bitcoin', 'UNI': 'Bitcoin',
        'WAN': 'Bitcoin', 'XLM': 'Bitcoin', 'XMR': 'Bitcoin',
        'XTZ': 'Bitcoin', 'YFI': 'Bitcoin', 'BTC': 'Tether'
    }
    assert symbols_by_crypto_currency_tickers == {
        'AAVE': {'AAVE/BTC', 'AAVE/USDT'}, 'ADA': {'ADA/BTC', 'ADA/USDT'}, 'ATOM': {'ATOM/BTC', 'ATOM/USDT'},
        'BAT': {'BAT/BTC', 'BAT/USDT'}, 'BNB': {'BNB/USDT', 'BNB/BTC'}, 'DASH': {'DASH/USDT', 'DASH/BTC'},
        'DOT': {'DOT/BTC', 'DOT/USDT'}, 'EOS': {'EOS/BTC', 'EOS/USDT'}, 'ETC': {'ETC/USDT', 'ETC/BTC'},
        'ETH': {'ETH/USDT', 'ETH/BTC'}, 'FIL': {'FIL/BTC', 'FIL/USDT'}, 'LINK': {'LINK/USDT', 'LINK/BTC'},
        'LTC': {'LTC/USDT', 'LTC/BTC'}, 'NEO': {'NEO/USDT', 'NEO/BTC'}, 'ONT': {'ONT/BTC', 'ONT/USDT'},
        'ROSE': {'ROSE/USDT', 'ROSE/BTC'}, 'SUSHI': {'SUSHI/USDT', 'SUSHI/BTC'}, 'SXP': {'SXP/BTC', 'SXP/USDT'},
        'THETA': {'THETA/USDT', 'THETA/BTC'}, 'TOMO': {'TOMO/BTC', 'TOMO/USDT'}, 'UNI': {'UNI/BTC', 'UNI/USDT'},
        'WAN': {'WAN/BTC', 'WAN/USDT'}, 'XLM': {'XLM/USDT', 'XLM/BTC'}, 'XMR': {'XMR/BTC', 'XMR/USDT'},
        'XTZ': {'XTZ/BTC', 'XTZ/USDT'}, 'YFI': {'YFI/USDT', 'YFI/BTC'}, 'BTC': {'BTC/USDT'}
    }


async def test_filter_pairs():
    exchange_api = ExchangeAPIMock()
    exchange_manager = None

    assert _filter_pairs(['BAT/BTC', 'BAT/USDT', 'BNB/USDT', 'BNB/BTC'], 'BAT', exchange_api, exchange_manager) == \
           {'BAT/BTC', 'BAT/USDT'}

    assert _filter_pairs(['BAT/BTC', 'BAT/USDT', 'BNB/USDT', 'BNB/BTC'], 'BNB', exchange_api, exchange_manager) == \
           {'BNB/BTC', 'BNB/USDT'}

    assert _filter_pairs(['BAT/BTC', 'BAT/USDT', 'BNB/USDT', 'BNB/BTC'], 'BTC', exchange_api, exchange_manager) == \
           set()

    assert _filter_pairs(['BAT/BTC', 'BAT/USDT', 'BNB/USDT', 'BNB/BTC'], 'USDT', exchange_api, exchange_manager) == \
           set()

    assert _filter_pairs([], 'USDT', exchange_api, exchange_manager) == \
           set()

    assert _filter_pairs([], 'USDT', exchange_api, exchange_manager) == \
           set()


class ExchangeAPIMock:

    def get_exchange_id_from_matrix_id(self, *args):
        return "1"

    def get_exchange_manager_from_exchange_name_and_id(self, *args):
        return None

    def get_base_currency(self, exchange_manager, symbol):
        return symbol_util.parse_symbol(symbol).base



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_commons.enums import PriceIndexes, TimeFrames
from octobot_evaluators.util.evaluation_util import get_eval_time, get_shortest_time_frame


def test_get_eval_time():
    candle = []
    candle.insert(PriceIndexes.IND_PRICE_TIME.value, 123456)
    partial_candle = []
    partial_candle.insert(PriceIndexes.IND_PRICE_TIME.value, 1234567)
    kline = []
    kline.insert(PriceIndexes.IND_PRICE_TIME.value, 12345678)
    assert get_eval_time(full_candle=candle, time_frame=TimeFrames.ONE_HOUR) == 123456 + 60 * 60
    assert get_eval_time(partial_candle=partial_candle) == 1234567
    assert get_eval_time(kline=kline) == 12345678


def test_get_shortest_time_frame():
    pref_time_frames = [TimeFrames.ONE_HOUR, TimeFrames.ONE_DAY, TimeFrames.ONE_MONTH]
    assert get_shortest_time_frame(TimeFrames.ONE_HOUR, pref_time_frames, []) == TimeFrames.ONE_HOUR
    assert get_shortest_time_frame(TimeFrames.ONE_MINUTE, pref_time_frames, [TimeFrames.ONE_WEEK]) == TimeFrames.ONE_HOUR
    assert get_shortest_time_frame(TimeFrames.ONE_MINUTE, [], [TimeFrames.ONE_MONTH]) == TimeFrames.ONE_MONTH
    assert get_shortest_time_frame(TimeFrames.ONE_MINUTE, [], [TimeFrames.ONE_MONTH, TimeFrames.ONE_DAY]) == TimeFrames.ONE_DAY
    assert get_shortest_time_frame(TimeFrames.ONE_MINUTE, [], [TimeFrames.ONE_HOUR, TimeFrames.ONE_MONTH]) == TimeFrames.ONE_HOUR



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.evaluators_util as evaluators_util
import octobot_commons.logging as logging

import octobot_evaluators.enums as enums
import octobot_evaluators.errors as errors
import octobot_evaluators.matrix as matrix


def get_matrix(matrix_id):
    """
    Get the matrix from its id
    :param matrix_id: the matrix id
    :return: the matrix instance
    """
    return matrix.Matrices.instance().get_matrix(matrix_id)


def set_tentacle_value(matrix_id, tentacle_path, tentacle_type, tentacle_value, timestamp=0):
    """
    Set the node value at tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle path
    :param tentacle_type: the tentacle type
    :param tentacle_value: the tentacle value
    :param timestamp: the value modification timestamp.
    """
    get_matrix(matrix_id).set_node_value(value=tentacle_value, value_type=tentacle_type,
                                         value_path=tentacle_path, timestamp=timestamp)


def get_tentacle_node(matrix_id, tentacle_path):
    """
    Return the node at tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle path
    :return: the tentacle node
    """
    return get_matrix(matrix_id).get_node_at_path(node_path=tentacle_path)


def delete_tentacle_node(matrix_id, tentacle_path):
    """
    Delete the node at tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle path
    :return: the deleted node
    """
    return get_matrix(matrix_id).delete_node_at_path(node_path=tentacle_path)


def get_tentacle_value(matrix_id, tentacle_path):
    """
    Get the value of the node at tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle path
    :return: the tentacle value
    """
    tentacle_node = get_tentacle_node(matrix_id, tentacle_path)
    if tentacle_node:
        return tentacle_node.node_value
    return None


def get_tentacle_eval_time(matrix_id, tentacle_path):
    """
    Get the evaluation time of the node at tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle path
    :return: the tentacle evaluation time
    """
    tentacle_node = get_tentacle_node(matrix_id, tentacle_path)
    if tentacle_node:
        return tentacle_node.node_value_time
    return None


def get_matrix_default_value_path(tentacle_name,
                                  tentacle_type,
                                  exchange_name=None,
                                  cryptocurrency=None,
                                  symbol=None,
                                  time_frame=None):
    """
    Create matrix value path with default path
    :param tentacle_name:
    :param tentacle_type:
    :param exchange_name:
    :param cryptocurrency:
    :param symbol:
    :param time_frame:
    :return: the default matrix
    """
    return get_tentacle_path(exchange_name=exchange_name,
                             tentacle_type=tentacle_type,
                             tentacle_name=tentacle_name) + get_tentacle_value_path(
        cryptocurrency=cryptocurrency,
        symbol=symbol,
        time_frame=time_frame)


def get_tentacle_nodes(matrix_id, exchange_name=None, tentacle_type=None, tentacle_name=None):
    """
    Returns the list of nodes related to the exchange_name, tentacle_type and tentacle_name, ignored if None
    :param matrix_id: the matrix id
    :param exchange_name: the exchange name to search for in the matrix
    :param tentacle_type: the tentacle type to search for in the matrix
    :param tentacle_name: the tentacle name to search for in the matrix
    :return: nodes linked to the given params
    """
    return get_matrix(matrix_id).get_node_children_at_path(get_tentacle_path(exchange_name=exchange_name,
                                                                             tentacle_type=tentacle_type,
                                                                             tentacle_name=tentacle_name))


def get_node_children_by_names_at_path(matrix_id, node_path, starting_node=None) -> dict:
    """
    :param matrix_id: the matrix id
    :param node_path: the node's path to inspect
    :param starting_node: the node to start the path from, default is the matrix root
    :return: a dict of the children nodes of the given path identified by their name
    """
    return get_matrix(matrix_id).get_node_children_by_names_at_path(node_path, starting_node=starting_node)


def get_tentacles_value_nodes(matrix_id, tentacle_nodes, cryptocurrency=None, symbol=None, time_frame=None):
    """
    Returns the list of nodes related to the symbol and / or time_frame from the given tentacle_nodes list
    :param matrix_id: the matrix id
    :param tentacle_nodes: the exchange name to search for in the matrix
    :param cryptocurrency: the cryptocurrency to search for in the given node list
    :param symbol: the symbol to search for in the given node list
    :param time_frame: the time frame to search for in the given nodes list
    :return: nodes linked to the given params
    """
    return [node_at_path for node_at_path in [
        get_matrix(matrix_id).get_node_at_path(get_tentacle_value_path(cryptocurrency=cryptocurrency,
                                                                       symbol=symbol,
                                                                       time_frame=time_frame),
                                               starting_node=n)
        for n in tentacle_nodes]
            if node_at_path is not None]


def get_latest_eval_time(matrix_id, exchange_name=None, tentacle_type=None, cryptocurrency=None,
                         symbol=None, time_frame=None):
    eval_times = []
    for value_node in matrix.get_tentacles_value_nodes(
            matrix_id,
            get_tentacle_nodes(matrix_id,
                               exchange_name=exchange_name,
                               tentacle_type=tentacle_type),
            cryptocurrency=cryptocurrency,
            symbol=symbol,
            time_frame=time_frame):

        if isinstance(value_node.node_value_time, (float, int)):
            eval_times.append(value_node.node_value_time)
    return max(eval_times) if eval_times else None


def get_tentacle_path(exchange_name=None, tentacle_type=None, tentacle_name=None) -> list:
    """
    Returns the path related to the tentacle name, type and exchange name
    :param tentacle_type: the tentacle type to add in the path, ignored if None
    :param tentacle_name: the tentacle name to add in the path, ignored if None
    :param exchange_name: the exchange name to add in the path (as the first element), ignored if None
    :return: a list of string that represents the path of the given params
    """
    node_path = []
    if exchange_name is not None:
        node_path.append(exchange_name)
    if tentacle_type is not None:
        node_path.append(tentacle_type)
    if tentacle_name is not None:
        node_path.append(tentacle_name)
    return node_path


def get_tentacle_value_path(cryptocurrency=None, symbol=None, time_frame=None) -> list:
    """
    Returns the path related to symbol and / or time_frame values
    :param cryptocurrency: the cryptocurrency to add in the path, ignored if None
    :param symbol: the symbol to add in the path, ignored if None
    :param time_frame: the time frame to add in the path, ignored if None
    :return: a list of string that represents the path of the given params
    """
    node_path: list = []
    if cryptocurrency is not None:
        node_path.append(cryptocurrency)
    if symbol is not None:
        node_path.append(symbol)
    if time_frame is not None:
        node_path.append(time_frame)
    return node_path


def get_evaluations_by_evaluator(matrix_id,
                                 exchange_name=None,
                                 tentacle_type=None,
                                 cryptocurrency=None,
                                 symbol=None,
                                 time_frame=None,
                                 allow_missing=True,
                                 allowed_values=None) -> dict:
    """
    Return a dict of evaluation nodes by evaluator name
    :param matrix_id: the matrix id
    :param exchange_name: the exchange name
    :param tentacle_type: the tentacle type
    :param cryptocurrency: the currency ticker
    :param symbol: the traded pair
    :param time_frame: the evaluation time frame
    :param allow_missing: if False will raise UnsetTentacleEvaluation on missing or invalid evaluation
    :param allowed_values: a white list of allowed values not to be taken as invalid
    :return: the dict of evaluation nodes by evaluator name
    """
    evaluator_nodes = get_node_children_by_names_at_path(matrix_id,
                                                         get_tentacle_path(exchange_name=exchange_name,
                                                                           tentacle_type=tentacle_type))
    evaluations_by_evaluator = {}
    for evaluator_name, node in evaluator_nodes.items():
        evaluation = get_tentacles_value_nodes(matrix_id, [node], cryptocurrency=cryptocurrency,
                                               symbol=symbol, time_frame=time_frame)
        if len(evaluation) > 1:
            logging.get_logger("matrix_manager").warning(
                "More than one evaluation corresponding to the given tentacle filter, "
                "this means there is an issue in this methods given arguments")
        elif evaluation:
            eval_value = evaluation[0].node_value
            if (allowed_values is not None and eval_value in allowed_values) or \
                    evaluators_util.check_valid_eval_note(eval_value):
                evaluations_by_evaluator[evaluator_name] = evaluation[0]
            elif not allow_missing:
                raise errors.UnsetTentacleEvaluation(f"Missing {time_frame if time_frame else 'evaluation'} "
                                                     f"for {evaluator_name} on {symbol}, evaluation is "
                                                     f"{repr(eval_value)}).")
    return evaluations_by_evaluator


def get_available_time_frames(matrix_id, exchange_name, tentacle_type, cryptocurrency, symbol) -> list:
    """
    Return the list of available time frames for the given tentacle
    :param matrix_id: the matrix id
    :param exchange_name: the exchange name
    :param tentacle_type: the tentacle type
    :param cryptocurrency: the currency ticker
    :param symbol: the traded pair
    :return: the list of available time frames for the given tentacle
    """
    try:
        evaluator_nodes = get_node_children_by_names_at_path(matrix_id,
                                                             get_tentacle_path(exchange_name=exchange_name,
                                                                               tentacle_type=tentacle_type))
        first_node = next(iter(evaluator_nodes.values()))
        return list(get_node_children_by_names_at_path(matrix_id,
                                                       get_tentacle_value_path(cryptocurrency=cryptocurrency,
                                                                               symbol=symbol),
                                                       starting_node=first_node))
    except StopIteration:
        return []


def get_available_symbols(matrix_id,
                          exchange_name,
                          cryptocurrency,
                          tentacle_type=enums.EvaluatorMatrixTypes.TA.value,
                          second_tentacle_type=enums.EvaluatorMatrixTypes.REAL_TIME.value) -> list:
    """
    Return the list of available symbols for the given currency
    :param matrix_id: the matrix id
    :param exchange_name: the exchange name
    :param cryptocurrency: the cryptocurrency ticker
    :param tentacle_type: the tentacle type to look into first
    :param second_tentacle_type: the tentacle type to look into if no symbol is found in the first tentacle type
    :return: the list of available symbols for the given currency
    """
    try:
        evaluator_nodes = get_node_children_by_names_at_path(matrix_id,
                                                             get_tentacle_path(exchange_name=exchange_name,
                                                                               tentacle_type=tentacle_type))
        first_node = next(iter(evaluator_nodes.values()))
        possible_symbols = list(get_node_children_by_names_at_path(
            matrix_id,
            get_tentacle_value_path(cryptocurrency=cryptocurrency),
            starting_node=first_node))
        if possible_symbols:
            return possible_symbols
        elif tentacle_type != second_tentacle_type:
            # try with second tentacle type
            return get_available_symbols(matrix_id, exchange_name,
                                         cryptocurrency, second_tentacle_type, second_tentacle_type)
    except StopIteration:
        return []


def is_tentacle_value_valid(matrix_id, tentacle_path, timestamp=0, delta=10) -> bool:
    """
    Check if the node is ready to be used
    WARNING: This method only works with complete default tentacle path
    :param matrix_id: the matrix id
    :param tentacle_path: the tentacle node path
    :param timestamp: the timestamp to use
    :param delta: the authorized delta to be valid (in seconds)
    :return: True if the node is valid else False
    """
    if timestamp == 0:
        timestamp = time.time()
    try:
        node = get_tentacle_node(matrix_id, tentacle_path)
        if node is None:
            raise KeyError(f"No node at {tentacle_path}")
        return timestamp - (node.node_value_time +
                            common_enums.TimeFramesMinutes[common_enums.TimeFrames(tentacle_path[-1])]
                            * common_constants.MINUTE_TO_SECONDS + delta) < 0
    except (IndexError, ValueError):
        return False


def is_tentacles_values_valid(matrix_id, tentacle_path_list, timestamp=0, delta=10) -> bool:
    """
    Check if each of the tentacle path value is valid
    :param matrix_id: the matrix id
    :param tentacle_path_list: the tentacle node path list
    :param timestamp: the timestamp to use
    :param delta: the authorized delta to be valid (in seconds)
    :return: True if all the node values are valid else False
    """
    return all([is_tentacle_value_valid(matrix_id=matrix_id,
                                        tentacle_path=tentacle_path,
                                        timestamp=timestamp,
                                        delta=delta)
                for tentacle_path in tentacle_path_list])



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_evaluators.constants as constants
import octobot_evaluators.enums as enums
import octobot_evaluators.evaluators as evaluator
import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.matrix as matrix

import octobot_tentacles_manager.api as api
import octobot_tentacles_manager.configuration as tm_configuration


class StrategyEvaluator(evaluator.AbstractEvaluator):
    __metaclass__ = evaluator.AbstractEvaluator

    def __init__(self, tentacles_setup_config):
        super().__init__(tentacles_setup_config)
        self.consumer_instance = None
        self.strategy_time_frames = []
        self.evaluations_last_updates = {}
        self.allowed_time_delta = None

        # caches
        self.available_evaluators_cache = {}
        self.available_time_frames_cache = {}
        self.available_node_paths_cache = {}

    def init_user_inputs(self, inputs: dict) -> None:
        """
        Called right before starting the tentacle, should define all the tentacle's user inputs unless
        those are defined somewhere else.
        """
        self.UI.user_input(constants.STRATEGIES_REQUIRED_TIME_FRAME, common_enums.UserInputTypes.MULTIPLE_OPTIONS,
                           [common_enums.TimeFrames.ONE_HOUR.value],
                           inputs, options=[tf.value for tf in common_enums.TimeFrames],
                           title="Time frame: The time frame to observe in order to spot changes.")

    async def start(self, bot_id: str) -> bool:
        """
        Default Strategy start: to be overwritten
        Subscribe to Matrix notification from self.symbols and self.time_frames
        :return: success of the evaluator's start
        """
        await super().start(bot_id)
        self.consumer_instance = await evaluator_channels.get_chan(constants.MATRIX_CHANNEL,
                                                                   self.matrix_id).new_consumer(
            self.strategy_matrix_callback,
            priority_level=self.priority_level,
            exchange_name=self.exchange_name if self.exchange_name else common_constants.CHANNEL_WILDCARD)
        self._init_exchange_allowed_time_delta(self.exchange_name, self.matrix_id)
        return True

    async def strategy_completed(self,
                                 cryptocurrency: str = None,
                                 symbol: str = None,
                                 time_frame=None,
                                 eval_note=None,
                                 eval_time=0,
                                 notify=True) -> None:
        """
        Main async method to notify that a strategy has updated its evaluation
        :param cryptocurrency: evaluated cryptocurrency
        :param symbol: evaluated symbol
        :param time_frame: evaluated time frame
        :param eval_note: if None = self.eval_note
        :param eval_time: the time of the evaluation if relevant, default is 0
        :param notify: if true, will trigger matrix consumers
        :return: None
        """
        return await self.evaluation_completed(cryptocurrency=cryptocurrency,
                                               symbol=symbol,
                                               time_frame=time_frame,
                                               eval_note=eval_note,
                                               eval_time=eval_time,
                                               notify=notify,
                                               origin_consumer=self.consumer_instance)

    def is_evaluator_cycle_complete(self, matrix_id, evaluator_name, evaluator_type, exchange_name,
                                    cryptocurrency, symbol, time_frame) -> bool:
        """
        :return: True if the strategy is to be waken up by evaluators of the given type at the moment of this call.
        This avoids partial time frame updates wakeup.
        Override if necessary
        """
        # 1. Ensure this evaluation has not already been sent
        # 2. Ensure every evaluator of the given type form this time frame are valid
        return not self._already_sent_this_technical_evaluation(matrix_id,
                                                                evaluator_name,
                                                                evaluator_type,
                                                                exchange_name,
                                                                cryptocurrency,
                                                                symbol,
                                                                time_frame) and \
               self._are_every_evaluation_valid_and_up_to_date(matrix_id,
                                                               evaluator_name,
                                                               evaluator_type,
                                                               exchange_name,
                                                               cryptocurrency,
                                                               symbol,
                                                               time_frame)

    def clear_cache(self):
        self.available_evaluators_cache = {}
        self.available_time_frames_cache = {}
        self.available_node_paths_cache = {}

    async def stop(self) -> None:
        await super().stop()
        if self.consumer_instance:
            await evaluator_channels.get_chan(constants.MATRIX_CHANNEL,
                                              self.matrix_id).remove_consumer(self.consumer_instance)
            self.consumer_instance = None

    def get_full_cycle_evaluator_types(self) -> tuple:
        # returns a tuple as it is faster to create than a list
        return enums.EvaluatorMatrixTypes.TA.value,

    async def strategy_matrix_callback(self,
                                       matrix_id,
                                       evaluator_name,
                                       evaluator_type,
                                       eval_note,
                                       eval_note_type,
                                       exchange_name,
                                       cryptocurrency,
                                       symbol,
                                       time_frame):
        # if this callback is from a technical evaluator: ensure strategy should be notified at this moment
        for full_cycle_evaluator in self.get_full_cycle_evaluator_types():
            if evaluator_type == full_cycle_evaluator:
                # ensure this time frame is within the strategy's time frames
                if common_enums.TimeFrames(time_frame) not in self.strategy_time_frames or \
                        not self.is_evaluator_cycle_complete(matrix_id,
                                                             evaluator_name,
                                                             evaluator_type,
                                                             exchange_name,
                                                             cryptocurrency,
                                                             symbol,
                                                             time_frame):
                    # do not call the strategy
                    return
        await self.matrix_callback(
            matrix_id,
            evaluator_name,
            evaluator_type,
            eval_note,
            eval_note_type,
            exchange_name,
            cryptocurrency,
            symbol,
            time_frame
        )

    async def matrix_callback(self,
                              matrix_id,
                              evaluator_name,
                              evaluator_type,
                              eval_note,
                              eval_note_type,
                              exchange_name,
                              cryptocurrency,
                              symbol,
                              time_frame):
        # To be used to trigger an evaluation
        # Do not forget to check if evaluator_name is self.name
        pass

    def _are_every_evaluation_valid_and_up_to_date(self,
                                                   matrix_id,
                                                   evaluator_name,
                                                   evaluator_type,
                                                   exchange_name,
                                                   cryptocurrency,
                                                   symbol,
                                                   time_frame,
                                                   can_retry=True):
        to_validate_node_paths = self._get_available_node_paths(matrix_id,
                                                                evaluator_type,
                                                                exchange_name,
                                                                cryptocurrency,
                                                                symbol,
                                                                use_cache=True)
        current_time = self._get_exchange_current_time(exchange_name, matrix_id)
        # ensure all evaluations are valid (do not trigger on an expired evaluation)
        try:
            if all(matrix.is_tentacle_value_valid(self.matrix_id, evaluation_node_path,
                                                  timestamp=current_time,
                                                  delta=self.allowed_time_delta)
                   for evaluation_node_path in to_validate_node_paths):
                self._save_last_evaluation(matrix_id, exchange_name, evaluator_type, evaluator_name,
                                           cryptocurrency, symbol, time_frame)
                return True
            return False
        except KeyError:
            self.clear_cache()
            if can_retry:
                return self._are_every_evaluation_valid_and_up_to_date(matrix_id,
                                                                       evaluator_name,
                                                                       evaluator_type,
                                                                       exchange_name,
                                                                       cryptocurrency,
                                                                       symbol,
                                                                       time_frame,
                                                                       can_retry=False)
            raise

    def _get_available_node_paths(self,
                                  matrix_id,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  use_cache=True):
        if use_cache:
            try:
                return self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency][symbol]
            except KeyError:
                # No cache usage here to be use to refresh data
                node_paths = self._inner_get_available_node_paths(matrix_id, evaluator_type, exchange_name,
                                                                  cryptocurrency, symbol, use_cache=False)
                if matrix_id not in self.available_node_paths_cache:
                    self.available_node_paths_cache[matrix_id] = {}
                if exchange_name not in self.available_node_paths_cache[matrix_id]:
                    self.available_node_paths_cache[matrix_id][exchange_name] = {}
                if evaluator_type not in self.available_node_paths_cache[matrix_id][exchange_name]:
                    self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type] = {}
                if cryptocurrency not in self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type]:
                    self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency] = {}
                self.available_node_paths_cache[matrix_id][exchange_name][evaluator_type][cryptocurrency][symbol] = \
                    node_paths
                return node_paths
        return self._inner_get_available_node_paths(matrix_id, evaluator_type, exchange_name,
                                                    cryptocurrency, symbol, use_cache=use_cache)

    def _inner_get_available_node_paths(self, matrix_id, evaluator_type, exchange_name, cryptocurrency, symbol,
                                        use_cache=True):
        paths = []
        for time_frame in self.get_available_time_frames(matrix_id, exchange_name, evaluator_type,
                                                         cryptocurrency, symbol, use_cache=use_cache):
            if common_enums.TimeFrames(time_frame) in self.strategy_time_frames:
                for evaluator_name in self._get_available_evaluators(matrix_id, exchange_name, evaluator_type,
                                                                     use_cache=use_cache):
                    path = matrix.get_matrix_default_value_path(tentacle_name=evaluator_name,
                                                                tentacle_type=evaluator_type,
                                                                exchange_name=exchange_name,
                                                                cryptocurrency=cryptocurrency,
                                                                symbol=symbol,
                                                                time_frame=time_frame)
                    if matrix.get_tentacle_node(matrix_id, path) is not None:
                        paths.append(path)
        return paths

    def _save_last_evaluation(self, matrix_id, exchange_name, evaluator_type, tentacle_name,
                              cryptocurrency, symbol, time_frame):
        self._set_last_evaluation_time(exchange_name,
                                       evaluator_type,
                                       cryptocurrency,
                                       symbol,
                                       time_frame,
                                       matrix.get_tentacle_eval_time(matrix_id,
                                                                     matrix.get_matrix_default_value_path(
                                                                         tentacle_name=tentacle_name,
                                                                         tentacle_type=evaluator_type,
                                                                         exchange_name=exchange_name,
                                                                         cryptocurrency=cryptocurrency,
                                                                         symbol=symbol,
                                                                         time_frame=time_frame)
                                                                     )
                                       )

    def _already_sent_this_technical_evaluation(self, matrix_id, evaluator, evaluator_type, exchange_name,
                                                cryptocurrency, symbol, time_frame):
        try:
            update_time = matrix.get_tentacle_eval_time(matrix_id,
                                                        matrix.get_matrix_default_value_path(
                                                            tentacle_name=evaluator,
                                                            tentacle_type=evaluator_type,
                                                            exchange_name=exchange_name,
                                                            cryptocurrency=cryptocurrency,
                                                            symbol=symbol,
                                                            time_frame=time_frame)
                                                        )
            return self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol][time_frame] \
                   == update_time
        except KeyError:
            return False

    def _set_last_evaluation_time(self, exchange_name, evaluator_type, cryptocurrency, symbol, time_frame, value):
        try:
            self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol][time_frame] = value
        except KeyError:
            if exchange_name not in self.evaluations_last_updates:
                self.evaluations_last_updates[exchange_name] = {}
            if evaluator_type not in self.evaluations_last_updates[exchange_name]:
                self.evaluations_last_updates[exchange_name][evaluator_type] = {}
            if cryptocurrency not in self.evaluations_last_updates[exchange_name][evaluator_type]:
                self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency] = {}
            if symbol not in self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency]:
                self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol] = {}
            self.evaluations_last_updates[exchange_name][evaluator_type][cryptocurrency][symbol] = {
                time_frame: value
            }

    def _get_exchange_current_time(self, exchange_name, matrix_id):
        try:
            import octobot_trading.api as exchange_api
            exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
                exchange_name,
                exchange_api.get_exchange_id_from_matrix_id(exchange_name, matrix_id)
            )
            return exchange_api.get_exchange_current_time(exchange_manager)
        except ImportError:
            self.logger.error("Strategy requires OctoBot-Trading package installed")

    def _init_exchange_allowed_time_delta(self, exchange_name, matrix_id):
        try:
            import octobot_trading.api as exchange_api
            exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(
                exchange_name,
                exchange_api.get_exchange_id_from_matrix_id(exchange_name, matrix_id)
            )
            self.allowed_time_delta = exchange_api.get_exchange_allowed_time_lag(exchange_manager)
        except ImportError:
            self.logger.error("Strategy requires OctoBot-Trading package installed")

    async def evaluators_callback(self,
                                  matrix_id,
                                  evaluator_name,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  time_frame,
                                  data):
        # Used to communicate between evaluators
        if data[constants.EVALUATOR_CHANNEL_DATA_ACTION] == constants.RESET_EVALUATION:
            for time_frame in data[constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES]:
                self._set_last_evaluation_time(exchange_name, enums.EvaluatorMatrixTypes.TA.value,
                                               cryptocurrency, symbol, time_frame.value, None)

    def _get_available_evaluators(self, matrix_id, exchange_name, tentacle_type, use_cache=True):
        if use_cache:
            try:
                return self.available_evaluators_cache[matrix_id][exchange_name][tentacle_type]
            except KeyError:
                available_evaluators = matrix.get_node_children_by_names_at_path(
                    matrix_id, matrix.get_tentacle_path(exchange_name=exchange_name, tentacle_type=tentacle_type)
                ).keys()
                if matrix_id not in self.available_evaluators_cache:
                    self.available_evaluators_cache[matrix_id] = {}
                if exchange_name not in self.available_evaluators_cache[matrix_id]:
                    self.available_evaluators_cache[matrix_id][exchange_name] = {}
                self.available_evaluators_cache[matrix_id][exchange_name][tentacle_type] = available_evaluators
                return available_evaluators
        return matrix.get_node_children_by_names_at_path(
            matrix_id, matrix.get_tentacle_path(exchange_name=exchange_name, tentacle_type=tentacle_type)
        ).keys()

    def get_available_time_frames(self,
                                  matrix_id,
                                  exchange_name=None,
                                  tentacle_type=None,
                                  cryptocurrency=None,
                                  symbol=None,
                                  use_cache=True):
        if use_cache:
            try:
                return self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency][symbol]
            except KeyError:
                available_time_frames = matrix.get_available_time_frames(
                    matrix_id, exchange_name, tentacle_type, cryptocurrency, symbol)
                if matrix_id not in self.available_time_frames_cache:
                    self.available_time_frames_cache[matrix_id] = {}
                if exchange_name not in self.available_time_frames_cache[matrix_id]:
                    self.available_time_frames_cache[matrix_id][exchange_name] = {}
                if tentacle_type not in self.available_time_frames_cache[matrix_id][exchange_name]:
                    self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type] = {}
                if cryptocurrency not in self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type]:
                    self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency] = {}
                self.available_time_frames_cache[matrix_id][exchange_name][tentacle_type][cryptocurrency][symbol] = \
                    available_time_frames
                return available_time_frames
        return matrix.get_available_time_frames(matrix_id, exchange_name, tentacle_type, cryptocurrency, symbol)

    def _get_tentacle_registration_topic(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames):
        strategy_currencies, symbols, self.strategy_time_frames = super()._get_tentacle_registration_topic(
            all_symbols_by_crypto_currencies,
            time_frames,
            real_time_time_frames)
        # by default no time frame registration for strategies
        return strategy_currencies, symbols, [self.time_frame]

    @classmethod
    def get_required_time_frames(cls, config: dict,
                                 tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration):
        if constants.CONFIG_FORCED_TIME_FRAME in config:
            return time_frame_manager.parse_time_frames(config[constants.CONFIG_FORCED_TIME_FRAME])
        strategy_config: dict = api.get_tentacle_config(tentacles_setup_config, cls)
        if constants.STRATEGIES_REQUIRED_TIME_FRAME in strategy_config:
            return time_frame_manager.parse_time_frames(strategy_config[constants.STRATEGIES_REQUIRED_TIME_FRAME])
        else:
            raise Exception(f"'{constants.STRATEGIES_REQUIRED_TIME_FRAME}' is missing in configuration file")

    @classmethod
    def get_required_evaluators(cls, tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                                strategy_config: dict = None) -> list:
        """
        :param tentacles_config: the tentacles config to find the current strategy config from
        :param strategy_config: the strategy configuration dict
        :return: the list of required evaluators, [CONFIG_WILDCARD] means any evaluator
        """
        strategy_config: dict = strategy_config or api.get_tentacle_config(tentacles_config, cls)
        if constants.STRATEGIES_REQUIRED_EVALUATORS in strategy_config:
            return strategy_config[constants.STRATEGIES_REQUIRED_EVALUATORS]
        else:
            raise Exception(f"'{constants.STRATEGIES_REQUIRED_EVALUATORS}' is missing in configuration file")

    @classmethod
    def get_compatible_evaluators_types(cls, tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                                        strategy_config: dict = None) -> list:
        """
        :param tentacles_config: the tentacles config to find the current strategy config from
        :param strategy_config: the strategy configuration dict
        :return: the list of compatible evaluator type, [CONFIG_WILDCARD] means any type
        """
        strategy_config: dict = strategy_config or api.get_tentacle_config(tentacles_config, cls)
        if constants.STRATEGIES_COMPATIBLE_EVALUATOR_TYPES in strategy_config:
            return strategy_config[constants.STRATEGIES_COMPATIBLE_EVALUATOR_TYPES]
        return [common_constants.CONFIG_WILDCARD]

    @classmethod
    def get_default_evaluators(cls, tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                               strategy_config: dict = None):
        strategy_config: dict = strategy_config or api.get_tentacle_config(tentacles_config, cls)
        if constants.TENTACLE_DEFAULT_CONFIG in strategy_config:
            return strategy_config[constants.TENTACLE_DEFAULT_CONFIG]
        else:
            required_evaluators = cls.get_required_evaluators(tentacles_config, strategy_config)
            if required_evaluators == common_constants.CONFIG_WILDCARD:
                return []
            return required_evaluators



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.constants as channel_constants

import octobot_commons.constants as common_constants
import octobot_commons.enums as common_enums
import octobot_commons.channels_name as channels_name
import octobot_commons.tree as commons_tree

import octobot_evaluators.constants as constants
import octobot_evaluators.evaluators as evaluator


class TAEvaluator(evaluator.AbstractEvaluator):
    __metaclass__ = evaluator.AbstractEvaluator
    DEFAULT_LIVE_PRICE_INIT_TIMEOUT = 5 * common_constants.MINUTE_TO_SECONDS

    def __init__(self, tentacles_setup_config):
        super().__init__(tentacles_setup_config)
        self.time_frame = None

        # True when this evaluator is only triggered on closed candles
        self.is_triggered_after_candle_close = True

        self._price_init_timeout = self.DEFAULT_LIVE_PRICE_INIT_TIMEOUT

    async def start(self, bot_id: str) -> bool:
        """
        Default TA start: to be overwritten
        Subscribe to OHLCV notification from self.symbols and self.time_frames
        :return: success of the evaluator's start
        """
        await super().start(bot_id)
        try:
            import octobot_trading.exchange_channel as exchanges_channel
            import octobot_trading.api as exchange_api
            exchange_id = exchange_api.get_exchange_id_from_matrix_id(self.exchange_name, self.matrix_id)
            time_frame_filter = [tf.value
                                 for tf in exchange_api.get_exchange_available_required_time_frames(
                                    self.exchange_name, exchange_id)]
            if len(time_frame_filter) == 1:
                time_frame_filter = time_frame_filter[0]
            await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value, exchange_id).\
                new_consumer(
                    self.evaluator_ohlcv_callback,
                    cryptocurrency=self.cryptocurrency if self.cryptocurrency else channel_constants.CHANNEL_WILDCARD,
                    symbol=self.symbol if self.symbol else channel_constants.CHANNEL_WILDCARD,
                    time_frame=self.time_frame.value if self.time_frame else time_frame_filter,
                    priority_level=self.priority_level,
            )
            if exchange_api.get_is_backtesting(
                    exchange_api.get_exchange_manager_from_exchange_name_and_id(self.exchange_name, exchange_id)
            ):
                self.use_backtesting_init_timeout()
            return True
        except ImportError as e:
            self.logger.error("Can't connect to OHLCV trading channel")
        return False

    def use_backtesting_init_timeout(self):
        self._price_init_timeout = 0

    async def reset_evaluation(self, cryptocurrency, symbol, time_frame):
        self.eval_note = common_constants.START_PENDING_EVAL_NOTE
        await self.evaluation_completed(cryptocurrency, symbol, time_frame, eval_time=0, notify=False)

    async def ohlcv_callback(self, exchange: str, exchange_id: str,
                             cryptocurrency: str, symbol: str, time_frame, candle, inc_in_construction_data):
        # To be used to trigger an evaluation when a new candle in closed or a re-evaluation is required
        pass

    async def evaluator_ohlcv_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                                       time_frame: str, candle: dict):
        if not self.get_is_symbol_wildcard():
            # do not wait for price in symbol wildcard or it will prevent evaluations for all
            # symbols waiting in this consumer queue
            await commons_tree.EventProvider.instance().wait_for_event(
                self.bot_id,
                commons_tree.get_exchange_path(
                    exchange,
                    common_enums.InitializationEventExchangeTopics.PRICE.value,
                    symbol=symbol,
                ),
                self._price_init_timeout
            )
        await self.ohlcv_callback(exchange, exchange_id, cryptocurrency, symbol, time_frame, candle, False)

    async def evaluators_callback(self,
                                  matrix_id,
                                  evaluator_name,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  time_frame,
                                  data):
        # Used to communicate between evaluators
        # ignore time frames related to other instances if case of a non time frame wildcard evaluator
        time_frames_to_update = data.get(constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES, []) \
            if self.get_is_time_frame_wildcard() else \
            [self.time_frame] if self.time_frame in data.get(constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES, []) else []
        if data[constants.EVALUATOR_CHANNEL_DATA_ACTION] == constants.TA_RE_EVALUATION_TRIGGER_UPDATED_DATA:
            try:
                import octobot_trading.api as exchange_api
                exchange_id = data[constants.EVALUATOR_CHANNEL_DATA_EXCHANGE_ID]
                symbol_data = self.get_exchange_symbol_data(exchange_name, exchange_id, symbol)
                for time_frame in time_frames_to_update:
                    candles_data = exchange_api.get_symbol_historical_candles(symbol_data, time_frame, limit=1)
                    # do not trigger ohlcv_callback on empty candles data
                    if all(values for values in candles_data.values()):
                        last_full_candle = exchange_api.get_candle_as_list(candles_data)
                        await self.ohlcv_callback(exchange_name, exchange_id, cryptocurrency,
                                                  symbol, time_frame.value, last_full_candle, True)
            except ImportError:
                self.logger.error(f"Can't get OHLCV: requires OctoBot-Trading package installed")
        elif data[constants.EVALUATOR_CHANNEL_DATA_ACTION] == constants.RESET_EVALUATION:
            for time_frame in time_frames_to_update:
                await self.reset_evaluation(cryptocurrency, symbol, time_frame.value)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.util as channel_util

import octobot_tentacles_manager.api as api

import octobot_commons.constants as common_constants
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.constants as constants
import octobot_evaluators.evaluators as evaluator
import octobot_evaluators.util as util


def init_time_frames_from_strategies(config, tentacles_setup_config) -> None:
    config[common_constants.CONFIG_TIME_FRAME] = get_time_frames_from_strategies(config, tentacles_setup_config)


def get_time_frames_from_strategies(config, tentacles_setup_config) -> list:
    time_frame_list = set(
        time_frame
        for strategies_eval_class in get_activated_strategies_classes(tentacles_setup_config)
        for time_frame in get_time_frames_from_strategy(strategies_eval_class, config, tentacles_setup_config)
    )
    return time_frame_manager.sort_time_frames(list(time_frame_list))


def get_time_frames_from_strategy(strategy_class, config, tentacles_setup_config) -> list:
    return strategy_class.get_required_time_frames(config, tentacles_setup_config)


def init_required_candles_count_from_evaluators_and_strategies(config, tentacles_setup_config) -> None:
    candles_counts = [util.get_required_candles_count(tentacle_class, tentacles_setup_config)
                      for tentacle_class in get_activated_evaluators(tentacles_setup_config)]
    config[common_constants.CONFIG_TENTACLES_REQUIRED_CANDLES_COUNT] = max(candles_counts) if candles_counts \
        else common_constants.DEFAULT_IGNORED_VALUE


def get_activated_evaluators(tentacles_setup_config):
    return get_activated_TA_evaluators_classes(tentacles_setup_config) + \
        get_activated_scripted_evaluators_classes(tentacles_setup_config) + \
        get_activated_real_time_evaluators_classes(tentacles_setup_config) + \
        get_activated_social_evaluators_classes(tentacles_setup_config) + \
        get_activated_strategies_classes(tentacles_setup_config)


def get_activated_strategies_classes(tentacles_setup_config):
    return _get_activated_classes(tentacles_setup_config, evaluator.StrategyEvaluator)


def get_activated_TA_evaluators_classes(tentacles_setup_config):
    return _get_activated_classes(tentacles_setup_config, evaluator.TAEvaluator)


def get_activated_scripted_evaluators_classes(tentacles_setup_config):
    return _get_activated_classes(tentacles_setup_config, evaluator.ScriptedEvaluator)


def get_activated_real_time_evaluators_classes(tentacles_setup_config):
    return _get_activated_classes(tentacles_setup_config, evaluator.RealTimeEvaluator)


def get_activated_social_evaluators_classes(tentacles_setup_config):
    return _get_activated_classes(tentacles_setup_config, evaluator.SocialEvaluator)


def _get_activated_classes(tentacles_setup_config, parent_class):
    return [
        child_class
        for child_class in tentacles_management.get_all_classes_from_parent(parent_class)
        if api.is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config, child_class.get_name())
    ]


async def create_evaluator_channels(matrix_id: str, is_backtesting: bool = False) -> None:
    await channel_util.create_all_subclasses_channel(evaluator_channels.EvaluatorChannel,
                                                     evaluator_channels.set_chan,
                                                     is_synchronized=is_backtesting, matrix_id=matrix_id)


def del_evaluator_channels(matrix_id: str) -> None:
    evaluator_channels.del_chan(constants.MATRIX_CHANNEL, matrix_id)
    evaluator_channels.del_chan(constants.EVALUATORS_CHANNEL, matrix_id)


def matrix_channel_exists(matrix_id: str) -> bool:
    try:
        evaluator_channels.get_chan(constants.MATRIX_CHANNEL, matrix_id)
        return True
    except KeyError:
        return False



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_evaluators.api.evaluators import create_matrix
from octobot_evaluators.matrix.matrix import Matrix
from octobot_evaluators.matrix.matrices import Matrices


def cleanup_matrices():
    matrices = Matrices.instance()
    for m_id in list(matrices.matrices):
        Matrices.instance().del_matrix(m_id)


def test_default_matrices():
    cleanup_matrices()
    matrices = Matrices.instance()
    assert matrices.matrices == {}


@pytest.mark.asyncio
async def test_add_matrix():
    matrices = Matrices.instance()
    assert matrices.matrices == {}

    created_matrix: Matrix = Matrix()
    Matrices.instance().add_matrix(created_matrix)

    assert matrices.matrices != {}
    assert created_matrix.matrix_id in matrices.matrices


@pytest.mark.asyncio
async def test_get_matrix():
    matrix_id = create_matrix()

    assert Matrices.instance().get_matrix(matrix_id) is not None

    with pytest.raises(KeyError):
        assert Matrices.instance().get_matrix(matrix_id + "t") is None


@pytest.mark.asyncio
async def test_del_matrix():
    matrices = Matrices.instance()
    matrix_id = create_matrix()

    assert matrix_id in matrices.matrices
    Matrices.instance().del_matrix(matrix_id)
    assert matrix_id not in matrices.matrices



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

cpdef double get_eval_time(list full_candle=*, object time_frame=*, list partial_candle=*, list kline=*)
cpdef object get_shortest_time_frame(object ideal_time_frame, object preferred_available_time_frames, object others)
cpdef object local_trading_context(object evaluator, str symbol, object time_frame, object trigger_cache_timestamp,
                                   str cryptocurrency=*, str exchange=*, str exchange_id=*,
                                   object trigger_source=*, object trigger_value=*)
cpdef object local_cache_client(object evaluator, str symbol, object time_frame, str exchange_name=*)
cpdef int get_required_candles_count(object trading_mode_class, object tentacles_setup_config)



# cython: language_level=3
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_commons.tree as tree


cdef class Matrix:
    cdef public str matrix_id
    cdef public tree.BaseTree matrix

    cpdef void set_node_value(self, object value, object value_type, list value_path, double timestamp=*)
    cpdef list get_node_children_at_path(self, list node_path, tree.BaseTreeNode starting_node=*)
    cpdef dict get_node_children_by_names_at_path(self, list node_path, tree.BaseTreeNode starting_node=*)
    cpdef tree.BaseTreeNode get_node_at_path(self, list node_path, tree.BaseTreeNode starting_node=*)
    cpdef tree.BaseTreeNode delete_node_at_path(self, list node_path, tree.BaseTreeNode starting_node=*)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import importlib

import async_channel.constants as channel_constants

import octobot_commons.channels_name as channels_name
import octobot_commons.enums as commons_enums
import octobot_commons.constants as commons_constants
import octobot_commons.errors as commons_errors
import octobot_commons.databases as commons_databases

import octobot_evaluators.evaluators as evaluator
import octobot_evaluators.util as evaluators_util
import octobot_tentacles_manager.api as tentacles_manager_api


class ScriptedEvaluator(evaluator.AbstractEvaluator):
    __metaclass__ = evaluator.AbstractEvaluator
    EVALUATOR_SCRIPT_MODULE = None

    def __init__(self, tentacles_setup_config):
        super().__init__(tentacles_setup_config)
        self._script = None
        self._are_candles_initialized = False
        self._has_script_been_called_once = False

    def post_init(self, tentacles_setup_config):
        # add config folder to importable files to import the user script
        tentacles_manager_api.import_user_tentacles_config_folder(tentacles_setup_config)

    async def start(self, bot_id: str) -> bool:
        """
        Default TA start: to be overwritten
        Subscribe to OHLCV notification from self.symbols and self.time_frames
        :return: success of the evaluator's start
        """
        await super().start(bot_id)
        try:
            import octobot_trading.api as exchange_api
            exchange_id = exchange_api.get_exchange_id_from_matrix_id(self.exchange_name, self.matrix_id)
            trigger_time_frames = self.get_trigger_time_frames()
            time_frame_filter = [tf.value
                                 for tf in exchange_api.get_exchange_available_required_time_frames(
                                    self.exchange_name, exchange_id)
                                 if tf.value in trigger_time_frames or
                                 trigger_time_frames == commons_constants.CONFIG_WILDCARD
                                 ]
            if trigger_time_frames != commons_constants.CONFIG_WILDCARD and \
               len(time_frame_filter) < len(trigger_time_frames):
                missing_time_frames = [tf for tf in trigger_time_frames if tf not in time_frame_filter]
                self.logger.error(f"Missing timeframe to satisfy {trigger_time_frames} required time frames. "
                                  f"Please activate those timeframes {missing_time_frames}")
            if len(time_frame_filter) == 1:
                time_frame_filter = time_frame_filter[0]
            cryptocurrency = self.cryptocurrency if self.cryptocurrency else channel_constants.CHANNEL_WILDCARD
            symbol = self.symbol if self.symbol else channel_constants.CHANNEL_WILDCARD
            time_frame = self.time_frame.value if self.time_frame else time_frame_filter
            self.consumers += await self._register_on_channels(exchange_id, cryptocurrency, symbol, time_frame, bot_id)
            return True
        except ImportError:
            self.logger.error("Can't connect to trading channels")
        return False

    async def evaluator_ohlcv_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                                       time_frame: str, candle: dict):
        # add a full candle to time to get the real time
        trigger_time = candle[commons_enums.PriceIndexes.IND_PRICE_TIME.value] + \
                       commons_enums.TimeFramesMinutes[commons_enums.TimeFrames(time_frame)] * \
                       commons_constants.MINUTE_TO_SECONDS
        await self._call_script(exchange, exchange_id, cryptocurrency, symbol,
                                trigger_time,
                                commons_enums.TriggerSource.OHLCV.value,
                                time_frame=time_frame, candle=candle)

    async def evaluator_kline_callback(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                                       time_frame, kline: dict):
        await self._call_script(exchange, exchange_id, cryptocurrency, symbol,
                                kline[commons_enums.PriceIndexes.IND_PRICE_TIME.value],
                                commons_enums.TriggerSource.KLINE.value,
                                time_frame=time_frame, kline=kline)

    async def evaluator_manual_callback(self, context=None, ignore_cache=False, **kwargs):
        """
        Called when this evaluator is triggered from a manual call
        :param context: the calling script's context
        :param kwargs: unused parameters
        :return: the evaluation value
        """
        async with context.nested_call_context(self) as local_context:
            try:
                # Cache is initialized at the 1st call: since a new instance of the evaluator is
                # potentially created each time, use cache to figure out if it has been called already.
                # Since self._has_script_been_called_once is only used in the context of cached evaluators,
                # it' fine to have it False all the time when no cache is used.
                self._has_script_been_called_once = self.use_cache() and local_context.has_cache(
                    local_context.symbol,
                    local_context.time_frame,
                    config_name=local_context.config_name
                )
                return_value, from_cache = await self._get_cached_or_computed_value(
                    local_context, ignore_cache=ignore_cache
                )
                if return_value == commons_constants.DO_NOT_OVERRIDE_CACHE:
                    value, missing = await local_context.get_cached_value()
                    local_context.ensure_no_missing_cached_value(missing)
                    return value, None

                # Arriving here, we know the evaluation was completed, we would otherwise be in the except statement
                # We can now cache the value if necessary
                if not from_cache and self.use_cache() and return_value != commons_constants.DO_NOT_CACHE:
                    await local_context.set_cached_value(return_value, flush_if_necessary=True)
                return return_value, None
            except (commons_errors.MissingDataError, commons_errors.ExecutionAborted,
                    commons_errors.NoCacheValue) as e:
                self.logger.debug(f"Can't compute evaluator value: {e}")
                return commons_constants.DO_NOT_CACHE, e

    async def _inner_call_script(self, context):
        # always call the script at least once to save plotting statements
        await self._pre_script_call(context)
        computed_value = await self.get_script()(context)
        self._has_script_been_called_once = True
        return computed_value

    async def _get_cached_or_computed_value(self, context, ignore_cache=False):
        computed_value = None
        called = False
        is_value_missing = True
        if not self._has_script_been_called_once:
            if self.use_cache():
                # init necessary settings before initializing cache
                await self._pre_script_call(context)
                # init cache to be sure its initialized before any call
                context.init_cache()
            computed_value = await self._inner_call_script(context)
            called = True
        if not called:
            if not ignore_cache and self.use_cache():
                computed_value, is_value_missing = await context.get_cached_value()
            if is_value_missing and not called:
                computed_value = await self._inner_call_script(context)
        return computed_value, not is_value_missing

    async def _call_script(self, exchange: str, exchange_id: str, cryptocurrency: str, symbol: str,
                           trigger_cache_timestamp: float, trigger_source: str,
                           time_frame: str = None, candle: dict = None, kline: dict = None):
        self.last_call = (exchange, exchange_id, cryptocurrency, symbol, trigger_cache_timestamp,
                          trigger_source, time_frame, candle, kline)
        context = evaluators_util.local_trading_context(
            self, symbol, time_frame, trigger_cache_timestamp, cryptocurrency=cryptocurrency,
            exchange=exchange, exchange_id=exchange_id, trigger_source=trigger_source,
            trigger_value=candle or kline
        )
        try:
            import octobot_trading.api as trading_api
            if not self._are_candles_initialized:
                self._are_candles_initialized = trading_api.are_symbol_candles_initialized(context.exchange_manager,
                                                                                           symbol, time_frame)
                if not self._are_candles_initialized:
                    self.logger.debug(f"Waiting for candles to be initialized before calling script "
                                      f"for {symbol} {time_frame}")
                    return
            self.eval_note, from_cache = await self._get_cached_or_computed_value(context)
            eval_time = None
            if trigger_source == commons_enums.TriggerSource.OHLCV.value:
                eval_time = evaluators_util.get_eval_time(full_candle=candle, time_frame=time_frame)
            elif trigger_source == commons_enums.TriggerSource.KLINE.value:
                eval_time = evaluators_util.get_eval_time(partial_candle=kline)
            if eval_time is None:
                self.logger.error("Can't compute evaluation time, using exchange time")
                eval_time = trading_api.get_exchange_current_time(context.exchange_manager)
            await self.evaluation_completed(cryptocurrency, symbol, time_frame,
                                            eval_time=eval_time, cache_client=context,
                                            cache_if_available=not from_cache)
        except (commons_errors.MissingDataError, commons_errors.ExecutionAborted) as e:
            self.logger.debug(f"Script execution aborted: {e}")
            self.eval_note = commons_constants.DO_NOT_CACHE
        except ImportError:
            self.logger.exception(f"Error when importing octobot-trading")
        except Exception as e:
            self.logger.exception(f"Error when calling evaluation script: {e}", True, e)

    async def _pre_script_call(self, context):
        try:
            import octobot_trading.modes.script_keywords.basic_keywords as basic_keywords
            # Always register activation_topics use input to enable changing it from run metadata
            # (where user inputs are registered)
            activation_topic_values = [
                commons_enums.ActivationTopics.FULL_CANDLES.value,
                commons_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value
            ]
            activation_method = await basic_keywords.get_activation_topics(
                context,
                commons_enums.ActivationTopics.FULL_CANDLES.value,
                activation_topic_values
            )
            self.is_triggered_after_candle_close = activation_method == \
                commons_enums.ActivationTopics.FULL_CANDLES.value
        except ImportError:
            self.logger.error("Can't read octobot_trading scripting_library")

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not symbol dependant else False
        """
        return False

    @classmethod
    def get_is_time_frame_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not time_frame dependant else False
        """
        return False

    def register_script_module(self, script_module):
        self.__class__.EVALUATOR_SCRIPT_MODULE = script_module
        self._script = self.get_script_from_module(script_module)

    @staticmethod
    def get_script_from_module(module):
        return module.script

    def get_script(self):
        return self._script

    async def user_commands_callback(self, bot_id, subject, action, data) -> None:
        await super().user_commands_callback(bot_id, subject, action, data)
        if action in (commons_enums.UserCommands.RELOAD_SCRIPT.value, commons_enums.UserCommands.RELOAD_CONFIG.value):
            # also reload script on RELOAD_CONFIG
            await self._reload_script(bot_id)

    async def _reload_script(self, bot_id):
        importlib.reload(self.__class__.EVALUATOR_SCRIPT_MODULE)
        self.register_script_module(self.__class__.EVALUATOR_SCRIPT_MODULE)
        # reload config
        await self.reload_config(bot_id)
        if self.last_call:
            # todo cancel and restart live tasks
            # recall script with for are_data_initialized to false to re-write initial data
            run_data_db, symbol_db = self._get_run_and_symbol_dbs()
            time_frames = None if self.get_is_time_frame_wildcard() else (self.time_frame.value, )
            run_data_db.set_initialized_flags(False)
            symbol_db.set_initialized_flags(False, time_frames)
            self._has_script_been_called_once = False
            try:
                await self._call_script(*self.last_call)
            finally:
                await run_data_db.flush()
                run_data_db.set_initialized_flags(True)
                await symbol_db.flush()
                symbol_db.set_initialized_flags(True, time_frames)

    def _get_run_and_symbol_dbs(self):
        try:
            import octobot_trading.api as trading_api
            exchange_manager = trading_api.get_exchange_manager_from_exchange_name_and_id(
                self.exchange_name,
                trading_api.get_exchange_id_from_matrix_id(self.exchange_name, self.matrix_id)
            )
            bot_id = trading_api.get_bot_id(exchange_manager)
            provider = commons_databases.RunDatabasesProvider.instance()
            return provider.get_run_db(bot_id), provider.get_symbol_db(bot_id, self.exchange_name, self.symbol)
        except ImportError:
            self.logger.error("required OctoBot-trading to get a trading mode writer")
            raise

    async def _register_on_channels(self, exchange_id, cryptocurrency, symbol, time_frame, bot_id):
        consumers = []
        try:
            import octobot_trading.exchange_channel as exchanges_channel
            registration_topics = self._get_channel_registration()
            if channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value in registration_topics:
                consumers.append(
                    await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value,
                                                     exchange_id).
                        new_consumer(self.evaluator_ohlcv_callback, cryptocurrency=cryptocurrency,
                                     symbol=symbol, time_frame=time_frame, priority_level=self.priority_level)
                )
            if channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value in registration_topics:
                consumers.append(
                    await exchanges_channel.get_chan(channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value,
                                                     exchange_id). \
                        new_consumer(self.evaluator_kline_callback, cryptocurrency=cryptocurrency,
                                     symbol=symbol, time_frame=time_frame, priority_level=self.priority_level)
                )
        except ImportError:
            self.logger.warning("Can't connect to trading channels")
        return consumers

    def _get_channel_registration(self):
        TOPIC_TO_CHANNEL_NAME = {
            commons_enums.ActivationTopics.FULL_CANDLES.value:
                channels_name.OctoBotTradingChannelsName.OHLCV_CHANNEL.value,
            commons_enums.ActivationTopics.IN_CONSTRUCTION_CANDLES.value:
                channels_name.OctoBotTradingChannelsName.KLINE_CHANNEL.value,
        }
        registration_channels = []
        # Activate on full candles only by default (same as technical evaluators)
        topic = self.specific_config.get(commons_constants.CONFIG_ACTIVATION_TOPICS.replace(" ", "_"),
                                         commons_enums.ActivationTopics.FULL_CANDLES.value)
        try:
            registration_channels.append(TOPIC_TO_CHANNEL_NAME[topic])
        except KeyError:
            self.logger.error(f"Unknown registration topic: {topic}")
        return registration_channels



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_evaluators.evaluators.channel cimport evaluators
from octobot_evaluators.evaluators.channel cimport evaluator_channel

from octobot_evaluators.evaluators.channel.evaluators cimport (
    EvaluatorsChannelConsumer,
    EvaluatorsChannelProducer,
    EvaluatorsChannel,
)
from octobot_evaluators.evaluators.channel.evaluator_channel cimport (
    EvaluatorChannelConsumer,
    EvaluatorChannelSupervisedConsumer,
    EvaluatorChannelProducer,
    EvaluatorChannel,
    set_chan,
    get_evaluator_channels,
    del_evaluator_channel_container,
    get_chan,
    del_chan,
)

__all__ = [
    "EvaluatorsChannelConsumer",
    "EvaluatorChannelSupervisedConsumer",
    "EvaluatorsChannelProducer",
    "EvaluatorsChannel",
    "EvaluatorChannelConsumer",
    "EvaluatorChannelProducer",
    "EvaluatorChannel",
    "set_chan",
    "get_evaluator_channels",
    "del_evaluator_channel_container",
    "get_chan",
    "del_chan",
]



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_evaluators.matrix as matrix


def get_matrix(matrix_id) -> matrix.Matrix:
    return matrix.Matrices.instance().get_matrix(matrix_id)


def del_matrix(matrix_id) -> matrix.Matrix:
    return matrix.Matrices.instance().del_matrix(matrix_id)


def get_node_children_by_names(matrix) -> dict:
    return matrix.get_node_children_by_names_at_path([])


def get_children_list(matrix_node) -> dict:
    return matrix_node.children


def has_children(matrix_node) -> bool:
    return bool(matrix_node.children)


def get_value(matrix_node) -> object:
    return matrix_node.node_value


def get_type(matrix_node) -> object:
    return matrix_node.node_type


def get_time(matrix_node) -> object:
    return matrix_node.node_value_time



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channel_instances

import octobot_commons.constants as common_constants
import octobot_commons.logging as logging
import octobot_commons.tentacles_management as tentacles_management
import octobot_commons.time_frame_manager as time_frame_manager

import octobot_evaluators.api as api
import octobot_evaluators.constants as constants
import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.matrix as matrix
import octobot_evaluators.evaluators as evaluator

import octobot_tentacles_manager.api as tentacles_manager_api

LOGGER_NAME = "EvaluatorsAPI"


async def create_and_start_all_type_evaluators(tentacles_setup_config: object,
                                               matrix_id: str,
                                               exchange_name: str,
                                               bot_id: str,
                                               symbols_by_crypto_currencies: dict = None,
                                               symbols: list = None,
                                               time_frames: list = None,
                                               real_time_time_frames: list = None,
                                               relevant_evaluators=common_constants.CONFIG_WILDCARD,
                                               ) -> list:
    return await evaluator.create_and_start_all_type_evaluators(tentacles_setup_config=tentacles_setup_config,
                                                                matrix_id=matrix_id,
                                                                exchange_name=exchange_name,
                                                                bot_id=bot_id,
                                                                symbols_by_crypto_currencies=symbols_by_crypto_currencies,
                                                                symbols=symbols,
                                                                time_frames=time_frames,
                                                                real_time_time_frames=real_time_time_frames,
                                                                relevant_evaluators=relevant_evaluators)


def get_evaluator_classes_from_type(evaluator_type, tentacles_setup_config, activated_only=True) -> list:
    if activated_only:
        return [cls for cls in tentacles_management.get_all_classes_from_parent(
            evaluator.EvaluatorClassTypes[evaluator_type]) if cls.is_enabled(tentacles_setup_config, False)]
    return tentacles_management.get_all_classes_from_parent(evaluator.EvaluatorClassTypes[evaluator_type])


async def initialize_evaluators(config, tentacles_setup_config) -> None:
    """
    :param config: bot config
    :param tentacles_setup_config: tentacles configuration
    """
    _init_time_frames(config, tentacles_setup_config)
    # take evaluators and strategies candles requirements into account if any
    api.init_required_candles_count_from_evaluators_and_strategies(config, tentacles_setup_config)


def get_evaluators_time_frames(config) -> list:
    return time_frame_manager.get_config_time_frame(config)


def update_time_frames_config(evaluator_class, tentacles_setup_config, time_frames) -> None:
    config_update = {
        constants.STRATEGIES_REQUIRED_TIME_FRAME: [tf.value for tf in time_frames]
    }
    tentacles_manager_api.update_tentacle_config(
        tentacles_setup_config,
        evaluator_class,
        config_update
    )


def _init_time_frames(config, tentacles_setup_config):
    # Init time frames using enabled strategies
    api.init_time_frames_from_strategies(config, tentacles_setup_config)


def create_matrix() -> str:
    created_matrix: matrix.Matrix = matrix.Matrix()
    matrix.Matrices.instance().add_matrix(created_matrix)
    return created_matrix.matrix_id


async def stop_evaluator(evaluator) -> None:
    return await evaluator.stop()


async def stop_evaluator_channel(matrix_id, chan_name) -> None:
    try:
        await evaluator_channels.get_chan(chan_name, matrix_id).stop()
    except Exception as e:
        logging.get_logger(LOGGER_NAME).exception(e, True, f"Error when stopping evaluator channel {chan_name}: {e}")


async def stop_all_evaluator_channels(matrix_id) -> None:
    for channel in channel_instances.ChannelInstances.instance().channels[matrix_id]:
        await stop_evaluator_channel(matrix_id, channel)



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import pytest_asyncio

from async_channel.util.channel_creator import create_channel_instance
from octobot_evaluators.api.initialization import del_evaluator_channels
from octobot_evaluators.evaluators.channel.evaluator_channel import get_chan, set_chan, EvaluatorChannel
import octobot_commons.asyncio_tools as asyncio_tools

from tests import matrix_id

EVALUATOR_CHANNEL_NAME = "Evaluator"


async def evaluator_callback():
    pass


@pytest_asyncio.fixture
async def evaluator_channel(matrix_id):
    channel = None
    try:
        del_evaluator_channels(matrix_id)
        channel = await create_channel_instance(EvaluatorChannel, set_chan, matrix_id=matrix_id)
        yield matrix_id
    finally:
        if channel is not None:
            # gracefully stop channel
            await channel.stop()
        del_evaluator_channels(matrix_id)


@pytest.mark.asyncio
async def test_evaluator_channel_get_consumer_from_filters(evaluator_channel):
    consumer_1 = await get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel).new_consumer(evaluator_callback)
    consumer_2 = await get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel).new_consumer(evaluator_callback)
    consumer_3 = await get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel).new_consumer(evaluator_callback)
    assert get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel) \
               .get_consumer_from_filters({}) == [consumer_1, consumer_2, consumer_3]
    assert get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel) \
               .get_consumer_from_filters({}, origin_consumer=consumer_2) == [consumer_1, consumer_3]
    assert get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel) \
               .get_consumer_from_filters({}, origin_consumer=consumer_1) == [consumer_2, consumer_3]
    assert get_chan(EVALUATOR_CHANNEL_NAME, evaluator_channel) \
               .get_consumer_from_filters({}, origin_consumer=consumer_3) == [consumer_1, consumer_2]



# pylint: disable=E0203
#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import async_channel.channels as channels
import async_channel.enums as channel_enums
import async_channel.constants as channel_constants
import async_channel.consumer as consumers
import async_channel.producer as producers

import octobot_commons.logging as logging

import octobot_evaluators.constants as constants


class EvaluatorChannelConsumer(consumers.Consumer):
    """
    Consumer adapted for EvaluatorChannel
    """


class EvaluatorChannelSupervisedConsumer(consumers.SupervisedConsumer):
    """
    SupervisedConsumer adapted for EvaluatorChannel
    """


class EvaluatorChannelProducer(producers.Producer):
    """
    Producer adapted for EvaluatorChannel
    """


class EvaluatorChannel(channels.Channel):
    PRODUCER_CLASS = EvaluatorChannelProducer
    CONSUMER_CLASS = EvaluatorChannelConsumer
    DEFAULT_PRIORITY_LEVEL = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

    def __init__(self, matrix_id):
        super().__init__()
        self.matrix_id = matrix_id

    def get_consumer_from_filters(self, consumer_filters, origin_consumer=None) -> list:
        """
        Returns the instance filtered consumers list except origin_consumer if provided
        :param consumer_filters: the consumer filters dict
        :param origin_consumer: the consumer behind the call if any else None
        :return: the filtered consumer list
        """
        return [consumer
                for consumer in super(EvaluatorChannel, self).get_consumer_from_filters(consumer_filters)
                if origin_consumer is None or consumer is not origin_consumer]


def set_chan(chan, name) -> None:
    chan_name = chan.get_name() if name else name

    try:
        evaluator_chan = channels.ChannelInstances.instance().channels[chan.matrix_id]
    except KeyError:
        channels.ChannelInstances.instance().channels[chan.matrix_id] = {}
        evaluator_chan = channels.ChannelInstances.instance().channels[chan.matrix_id]

    if chan_name not in evaluator_chan:
        evaluator_chan[chan_name] = chan
    else:
        raise ValueError(f"Channel {chan_name} already exists.")


def get_evaluator_channels(matrix_id) -> dict:
    try:
        return channels.ChannelInstances.instance().channels[matrix_id]
    except KeyError as e:
        raise KeyError(f"Channels not found with matrix_id: {matrix_id}") from e


def del_evaluator_channel_container(matrix_id):
    try:
        channels.ChannelInstances.instance().channels.pop(matrix_id, None)
    except KeyError as e:
        raise KeyError(f"Channels not found with matrix_id: {matrix_id}") from e


def get_chan(chan_name, matrix_id) -> EvaluatorChannel:
    try:
        return channels.ChannelInstances.instance().channels[matrix_id][chan_name]
    except KeyError as e:
        raise KeyError(f"Channel {chan_name} not found with matrix_id: {matrix_id}") from e


def del_chan(chan_name, matrix_id) -> None:
    try:
        channels.ChannelInstances.instance().channels[matrix_id].pop(chan_name, None)
    except KeyError:
        logging.get_logger(EvaluatorChannel.__name__).warning(f"Can't del chan {chan_name} with matrix_id: {matrix_id}")


async def trigger_technical_evaluators_re_evaluation_with_updated_data(matrix_id,
                                                                       evaluator_name,
                                                                       evaluator_type,
                                                                       exchange_name,
                                                                       cryptocurrency,
                                                                       symbol,
                                                                       exchange_id,
                                                                       time_frames
                                                                       ):
    # first reset evaluations to avoid partially updated TA cycle validation
    await get_chan(constants.EVALUATORS_CHANNEL, matrix_id).get_internal_producer().send(
        matrix_id,
        data={
            constants.EVALUATOR_CHANNEL_DATA_ACTION: constants.RESET_EVALUATION,
            constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES: time_frames
        },
        evaluator_name=evaluator_name,
        evaluator_type=evaluator_type,
        exchange_name=exchange_name,
        cryptocurrency=cryptocurrency,
        symbol=symbol,
        time_frame=channel_constants.CHANNEL_WILDCARD
    )
    await get_chan(constants.EVALUATORS_CHANNEL, matrix_id).get_internal_producer().send(
        matrix_id,
        data={
            constants.EVALUATOR_CHANNEL_DATA_ACTION: constants.TA_RE_EVALUATION_TRIGGER_UPDATED_DATA,
            constants.EVALUATOR_CHANNEL_DATA_EXCHANGE_ID: exchange_id,
            constants.EVALUATOR_CHANNEL_DATA_TIME_FRAMES: time_frames
        },
        evaluator_name=evaluator_name,
        evaluator_type=evaluator_type,
        exchange_name=exchange_name,
        cryptocurrency=cryptocurrency,
        symbol=symbol,
        time_frame=channel_constants.CHANNEL_WILDCARD
    )



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_evaluators.matrix cimport matrix_manager
from octobot_evaluators.matrix cimport matrices
from octobot_evaluators.matrix cimport matrix

from octobot_evaluators.matrix.matrix_manager cimport (
    get_matrix,
    set_tentacle_value,
    get_tentacle_node,
    get_tentacle_value,
    get_tentacle_eval_time,
    get_matrix_default_value_path,
    get_tentacle_nodes,
    get_node_children_by_names_at_path,
    get_tentacles_value_nodes,
    get_latest_eval_time,
    get_tentacle_path,
    get_tentacle_value_path,
    get_evaluations_by_evaluator,
    get_available_time_frames,
    get_available_symbols,
    is_tentacle_value_valid,
    is_tentacles_values_valid,
)
from octobot_evaluators.matrix.matrices cimport (
    Matrices,
)
from octobot_evaluators.matrix.matrix cimport (
    Matrix,
)

from octobot_evaluators.matrix cimport channel
from octobot_evaluators.matrix.channel cimport (
    MatrixChannelConsumer,
    MatrixChannelSupervisedConsumer,
    MatrixChannelProducer,
    MatrixChannel,
)

__all__ = [
    "get_matrix",
    "set_tentacle_value",
    "get_tentacle_node",
    "get_tentacle_value",
    "get_tentacle_eval_time",
    "get_matrix_default_value_path",
    "get_tentacle_nodes",
    "get_node_children_by_names_at_path",
    "get_tentacles_value_nodes",
    "get_latest_eval_time",
    "get_tentacle_path",
    "get_tentacle_value_path",
    "get_evaluations_by_evaluator",
    "get_available_time_frames",
    "get_available_symbols",
    "is_tentacle_value_valid",
    "is_tentacles_values_valid",
    "Matrices",
    "Matrix",
    "MatrixChannelConsumer",
    "MatrixChannelSupervisedConsumer",
    "MatrixChannelProducer",
    "MatrixChannel",
]



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid

import octobot_commons.tree as tree


class Matrix:
    """
    Matrix dataclass store tentacles data in a BaseTree
    """
    __slots__ = ['matrix_id', 'matrix']

    def __init__(self):
        """
        Initialize the matrix as an BaseTree instance
        """
        self.matrix_id = str(uuid.uuid4())
        self.matrix = tree.BaseTree()

    def set_node_value(self, value, value_type, value_path, timestamp=0):
        """
        Set the node value at node path
        :param value_path: the node path
        :param value_type: the node type
        :param value: the node value
        :param timestamp: the value modification timestamp.
        """
        self.matrix.set_node_at_path(value, value_type, value_path, timestamp=timestamp)

    def get_node_children_at_path(self, node_path, starting_node=None):
        """
        Get the node children list
        :param node_path: the node path
        :param starting_node: the node to start the relative path
        :return: the list of node children
        """
        try:
            return list(self.matrix.get_node(node_path, starting_node=starting_node).children.values())
        except tree.NodeExistsError:
            return []

    def get_node_children_by_names_at_path(self, node_path, starting_node=None):
        """
        Get the node children dict with node name as key
        :param node_path: the node path
        :param starting_node: the node to start the relative path
        :return: the dict of node children
        """
        try:
            return {key: val
                    for key, val in self.matrix.get_node(node_path, starting_node=starting_node).children.items()}
        except tree.NodeExistsError:
            return {}

    def get_node_at_path(self, node_path, starting_node=None):
        """
        Get the BaseTreeNode at path
        :param node_path: the node path
        :param starting_node: the node to start the relative path
        :return: the node instance at path
        """
        try:
            return self.matrix.get_node(node_path, starting_node=starting_node)
        except tree.NodeExistsError:
            return None

    def delete_node_at_path(self, node_path, starting_node=None):
        """
        Delete the BaseTreeNode at path
        :param node_path: the node path
        :param starting_node: the node to start the relative path
        :return: the deleted node
        """
        try:
            return self.matrix.delete_node(node_path, starting_node=starting_node)
        except tree.NodeExistsError:
            return None



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as common_constants
import octobot_commons.tentacles_management as tentacles_management

import octobot_evaluators.evaluators as evaluator


def is_relevant_evaluator(evaluator_instance, relevant_evaluators, use_relevant_evaluators_only=False) -> bool:
    if evaluator_instance.enabled or use_relevant_evaluators_only:
        if relevant_evaluators == common_constants.CONFIG_WILDCARD or \
                evaluator_instance.get_name() in relevant_evaluators:
            return True
        else:
            parent_classes_names = [e.get_name() for e in evaluator_instance.get_parent_evaluator_classes()]
            to_check_set = relevant_evaluators
            if not isinstance(relevant_evaluators, set):
                to_check_set = set(relevant_evaluators)
            return not to_check_set.isdisjoint(parent_classes_names)
    return False


def get_relevant_TAs_for_strategy(strategy, tentacles_setup_config) -> list:
    ta_classes_list = []
    relevant_evaluators = strategy.get_required_evaluators(tentacles_setup_config)
    for ta_eval_class in tentacles_management.get_all_classes_from_parent(evaluator.TAEvaluator):
        ta_eval_class_instance = ta_eval_class(tentacles_setup_config)
        # use ony relevant_evaluators given by the strategy
        if common_constants.CONFIG_WILDCARD in relevant_evaluators or \
                is_relevant_evaluator(ta_eval_class_instance, relevant_evaluators, use_relevant_evaluators_only=True):
            ta_classes_list.append(ta_eval_class)
    return ta_classes_list



#  Drakkar-Software OctoBot-Evaluators
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import time

import octobot_tentacles_manager.api as api
import octobot_tentacles_manager.configuration as tm_configuration

import async_channel.constants as channel_constants
import async_channel.enums as channel_enums
import async_channel.channels as channels

import octobot_commons.constants as common_constants
import octobot_commons.errors as commons_errors
import octobot_commons.enums as commons_enums
import octobot_commons.tentacles_management as tentacles_management

import octobot_evaluators.evaluators.channel as evaluator_channels
import octobot_evaluators.constants as constants
import octobot_evaluators.matrix as matrix
import octobot_evaluators.evaluators.evaluator_factory as evaluator_factory

import octobot_evaluators.util as util


class AbstractEvaluator(tentacles_management.AbstractTentacle):
    __metaclass__ = tentacles_management.AbstractTentacle
    USER_INPUT_TENTACLE_TYPE = commons_enums.UserInputTentacleTypes.EVALUATOR

    def __init__(self, tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration):
        super().__init__()
        # Evaluator matrix id
        self.matrix_id: str = None

        # OctoBot id this evaluator has been started with
        self.bot_id: str = None

        # Tentacle global setup configuration
        self.tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration = tentacles_setup_config

        # Evaluator specific config (Is loaded from tentacle specific file)
        self.specific_config: dict = {}

        # If this indicator is enabled
        self.enabled: bool = self.is_enabled(self.tentacles_setup_config, False)

        # Specified Cryptocurrency for this instance (Should be None if wildcard)
        self.cryptocurrency: str = None

        # Specified Cryptocurrency name for this instance (Should be None if wildcard)
        self.cryptocurrency_name: str = None

        # Symbol is the cryptocurrency pair (Should be None if wildcard)
        self.symbol: str = None

        # Evaluation related exchange name
        self.exchange_name: str = None

        # Time_frame is the chart time frame (Should be None if wildcard)
        self.time_frame = None

        # history time represents the period of time of the indicator
        self.history_time = None

        # Evaluator category
        self.evaluator_type = None

        # Eval note will be set by the eval_impl at each call
        self.eval_note = common_constants.START_PENDING_EVAL_NOTE

        # Pertinence of indicator will be used with the eval_note to provide a relevancy
        self.pertinence = constants.START_EVAL_PERTINENCE

        # Active tells if this evaluator is currently activated (an evaluator can be paused)
        self.is_active: bool = True

        self.eval_note_time_to_live = None
        self.eval_note_changed_time = None

        # Define evaluators default consumer priority level
        self.priority_level: int = channel_enums.ChannelConsumerPriorityLevels.MEDIUM.value

        self.consumers = []

        # True when this evaluator is only triggered on closed candles
        self.is_triggered_after_candle_close = False

    def post_init(self, tentacles_setup_config):
        """
        Automatically called after __init__ when post_init is True (default) in evaluator_factory
        Override when necessary
        :param tentacles_setup_config: the tentacles_setup_config __init__ argument
        :return: None
        """
        pass

    @classmethod
    async def single_evaluation(
            cls,
            tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration,
            specific_config: dict,
            ignore_cache=False,
            should_trigger_post_init=False,
            **kwargs):
        evaluator_instance = evaluator_factory.create_temporary_evaluator_with_local_config(
            cls, tentacles_setup_config, specific_config, should_trigger_post_init)
        evaluation, error = await evaluator_instance.evaluator_manual_callback(ignore_cache=ignore_cache, **kwargs)
        return evaluation, error, evaluator_instance

    async def evaluator_manual_callback(self, **kwargs):
        """
        Override this method to define the appropriate behavior when this evaluator is being
        called manually
        :param kwargs: keyword arguments to be used for evaluation
        :return: the evaluation value
        """
        raise NotImplementedError("evaluator_manual_callback is not implemented")

    @staticmethod
    def get_eval_type():
        """
        Override this method when self.eval_note is other than : START_PENDING_EVAL_NOTE or float[-1:1]
        :return: type
        """
        return constants.EVALUATOR_EVAL_DEFAULT_TYPE

    @classmethod
    def get_is_cryptocurrencies_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not cryptocurrency dependant else False
        """
        return True

    @classmethod
    def get_is_cryptocurrency_name_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not cryptocurrency name dependant else False
        """
        return True

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not symbol dependant else False
        """
        return True

    @classmethod
    def get_is_time_frame_wildcard(cls) -> bool:
        """
        :return: True if the evaluator is not time_frame dependant else False
        """
        return True

    def get_trigger_time_frames(self):
        return self.specific_config.get(common_constants.CONFIG_TRIGGER_TIMEFRAMES, common_constants.CONFIG_WILDCARD)

    @staticmethod
    def invalidate_cache_on_code_change():
        # False is not yet supported
        return True

    @staticmethod
    def invalidate_cache_on_config_change():
        # False is not yet supported
        return True

    @classmethod
    def use_cache(cls):
        return False

    def get_local_config(self):
        return self.specific_config

    def _get_tentacle_registration_topic(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames):
        currencies = [self.cryptocurrency]
        symbols = [self.symbol]
        available_time_frames = [self.time_frame]
        if self.get_is_cryptocurrencies_wildcard():
            currencies = all_symbols_by_crypto_currencies.keys()
        if self.get_is_symbol_wildcard():
            symbols = [currency_symbol
                       for currency_symbols in all_symbols_by_crypto_currencies.values()
                       for currency_symbol in currency_symbols]
        if self.get_is_time_frame_wildcard():
            available_time_frames = time_frames
        trigger_timeframes = self.get_trigger_time_frames()
        if trigger_timeframes != common_constants.CONFIG_WILDCARD:
            available_time_frames = [tf
                                     for tf in available_time_frames
                                     if tf in trigger_timeframes]
        return currencies, symbols, available_time_frames

    async def initialize(self, all_symbols_by_crypto_currencies, time_frames, real_time_time_frames, bot_id):
        await self.reload_config(bot_id)
        currencies, symbols, time_frames = self._get_tentacle_registration_topic(all_symbols_by_crypto_currencies,
                                                                                 time_frames,
                                                                                 real_time_time_frames)
        for currency in currencies:
            for symbol in symbols:
                if symbol is None or symbol in all_symbols_by_crypto_currencies[currency]:
                    for time_frame in time_frames:
                        matrix.set_tentacle_value(
                            matrix_id=self.matrix_id,
                            tentacle_type=self.get_eval_type(),
                            tentacle_value=None,
                            tentacle_path=matrix.get_matrix_default_value_path(
                                exchange_name=self.exchange_name,
                                tentacle_type=self.evaluator_type.value,
                                tentacle_name=self.get_name(),
                                cryptocurrency=currency,
                                symbol=symbol,
                                time_frame=time_frame.value if time_frame else None
                            )
                        )

    async def evaluation_completed(self,
                                   cryptocurrency: str = None,
                                   symbol: str = None,
                                   time_frame=None,
                                   eval_note=None,
                                   eval_time=0,
                                   notify=True,
                                   origin_consumer=None,
                                   cache_client=None,
                                   cache_if_available=True) -> None:
        """
        Main async method to notify matrix to update
        :param cryptocurrency: evaluated cryptocurrency
        :param symbol: evaluated symbol
        :param time_frame: evaluated time frame
        :param eval_note: if None = self.eval_note
        :param eval_time: the time of the evaluation if relevant, default is 0
        :param notify: if true, will trigger matrix consumers
        :param origin_consumer: the sender consumer if it doesn't want to be notified
        :param cache_client: an existing cache client to avoid creating a local one
        :param cache_if_available: when True, if the evaluator is using cache, its value will be cached
        :return: None
        """
        try:
            if eval_note is None:
                eval_note = self.eval_note if self.eval_note is not None else common_constants.START_PENDING_EVAL_NOTE

            if self.use_cache():
                cache_client = cache_client or util.local_cache_client(self, symbol, time_frame)
                if self.eval_note == common_constants.DO_NOT_OVERRIDE_CACHE:
                    self.eval_note, missing = await cache_client.get_cached_value(cache_key=eval_time)
                    cache_client.ensure_no_missing_cached_value(missing)
                    eval_note = self.eval_note
                elif cache_if_available and eval_note != common_constants.DO_NOT_CACHE:
                    await cache_client.set_cached_value(eval_note, cache_key=eval_time, flush_if_necessary=True)
            self.ensure_eval_note_is_not_expired()
            await evaluator_channels.get_chan(constants.MATRIX_CHANNEL,
                                              self.matrix_id).get_internal_producer().send_eval_note(
                matrix_id=self.matrix_id,
                evaluator_name=self.get_name(),
                evaluator_type=self.evaluator_type.value,
                eval_note=eval_note,
                eval_note_type=self.get_eval_type(),
                eval_time=eval_time,
                exchange_name=self.exchange_name,
                cryptocurrency=cryptocurrency,
                symbol=symbol,
                time_frame=time_frame,
                notify=notify,
                origin_consumer=origin_consumer)
        except commons_errors.NoCacheValue:
            self.logger.warning(f"Evaluation as \"{common_constants.DO_NOT_OVERRIDE_CACHE}\" "
                                f"but the is no cache to publish an evaluation from")
        except Exception as e:
            # if ConfigManager.is_in_dev_mode(self.config): # TODO
            #     raise e
            # else:
            self.logger.exception(e, True, f"Exception in evaluation_completed(): {e}")
        finally:
            if self.eval_note == "nan":
                self.eval_note = common_constants.START_PENDING_EVAL_NOTE
                self.logger.warning(str(self.symbol) + " evaluator returned 'nan' as eval_note, ignoring this value.")

    async def start(self, bot_id: str) -> bool:
        """
        :return: success of the evaluator's start
        """
        self.bot_id = bot_id
        self.consumers.append(
            await evaluator_channels.get_chan(constants.EVALUATORS_CHANNEL, self.matrix_id).new_consumer(
                self.evaluators_callback,
                cryptocurrency=self.cryptocurrency if self.cryptocurrency else channel_constants.CHANNEL_WILDCARD,
                symbol=self.symbol if self.symbol else channel_constants.CHANNEL_WILDCARD,
                time_frame=self.time_frame if self.time_frame else channel_constants.CHANNEL_WILDCARD,
                priority_level=self.priority_level,
            )
        )

        try:
            import octobot_services.channel as services_channels
            self.consumers.append(
                await channels.get_chan(services_channels.UserCommandsChannel.get_name()).new_consumer(
                    self.user_commands_callback,
                    {"bot_id": bot_id, "subject": self.get_name()}
                )
            )
        except KeyError:
            # UserCommandsChannel might not be available
            pass
        except ImportError:
            self.logger.warning("Can't connect to services channels")

    async def user_commands_callback(self, bot_id, subject, action, data) -> None:
        self.logger.debug(f"Received {action} command")
        if action == commons_enums.UserCommands.RELOAD_CONFIG.value:
            await self.reload_config(bot_id)
            self.logger.debug("Reloaded configuration")

    async def stop(self) -> None:
        """
        implement if necessary
        :return: None
        """
        for consumer in self.consumers:
            await consumer.stop()

    async def prepare(self) -> None:
        """
        Called just before start(), implement if necessary
        :return: None
        """
        pass

    async def start_evaluator(self, bot_id: str) -> None:
        """
        Start a task as matrix producer
        :return: None
        """
        if await self.start(bot_id):
            self.logger.debug("Evaluator started")
        else:
            self.logger.debug("Evaluator not started")

    async def reload_config(self, bot_id: str) -> None:
        self.set_default_config()
        specific_config = api.get_tentacle_config(self.tentacles_setup_config, self.__class__)

        if not specific_config and self.ALLOW_SUPER_CLASS_CONFIG:
            # if nothing in config, try with any super-class' config file
            for super_class in self.get_parent_evaluator_classes(AbstractEvaluator):
                try:
                    if specific_config := api.get_tentacle_config(self.tentacles_setup_config, super_class):
                        break
                except KeyError:
                    pass  # super_class tentacle config not found
        self.specific_config.update(specific_config)
        await self.load_and_save_user_inputs(bot_id)
        self.logger.debug(f"Using config: {self.specific_config}")

    @classmethod
    def create_local_instance(cls, _, tentacles_setup_config, tentacle_config):
        return evaluator_factory.create_temporary_evaluator_with_local_config(
            cls, tentacles_setup_config, tentacle_config, False
        )

    def set_default_config(self):
        """
        To implement in subclasses if config necessary
        :return:
        """
        self.specific_config = {}

    @classmethod
    def get_evaluator_priority(cls, tentacles_setup_config) -> float:
        """
        Returns the priority of the evaluator (will later be compared to other evaluators).
        A higher priority evaluator will be called first when multiple evaluators are to
        be called at the same time.
        Default priority is DEFAULT_PRIORITY defined in OctoBot-Commons.
        Order is undefined between evaluators of the same priority.
        :return: the priority level
        """
        return api.get_tentacle_config(tentacles_setup_config, cls).get(common_constants.EVALUATOR_PRIORITY,
                                                                        common_constants.DEFAULT_EVALUATOR_PRIORITY)

    def reset(self) -> None:
        """
        Reset temporary parameters to enable fresh start
        :return: None
        """
        self.eval_note = common_constants.START_PENDING_EVAL_NOTE

    @classmethod
    def has_class_in_parents(cls, klass) -> bool:
        """
        Explore up to the 2nd degree parent
        :param klass: python Class to explore
        :return: Boolean
        """
        if klass in cls.__bases__:
            return True
        elif any(klass in base.__bases__ for base in cls.__bases__):
            return True
        else:
            for base in cls.__bases__:
                if any(klass in super_base.__bases__ for super_base in base.__bases__):
                    return True
        return False

    @classmethod
    def get_parent_evaluator_classes(cls, higher_parent_class_limit=None) -> list:
        """
        Return the evaluator parent classe(s)
        :param higher_parent_class_limit:
        :return: list of classes
        """
        return [
            class_type
            for class_type in cls.mro()
            if (higher_parent_class_limit if higher_parent_class_limit else AbstractEvaluator) in class_type.mro()
        ]

    def set_eval_note(self, new_eval_note) -> None:
        """
        Performs additionnal check to eval_note before changing it
        :param new_eval_note:
        :return: None
        """
        self.eval_note_changed()

        if self.eval_note == common_constants.START_PENDING_EVAL_NOTE:
            self.eval_note = common_constants.INIT_EVAL_NOTE

        if self.eval_note + new_eval_note > 1:
            self.eval_note = 1
        elif self.eval_note + new_eval_note < -1:
            self.eval_note = -1
        else:
            self.eval_note += new_eval_note

    @classmethod
    def is_enabled(cls, tentacles_setup_config, default) -> bool:
        """
        Check if the evaluator is enabled by configuration
        :param tentacles_setup_config: tentacles setup config
        :param default: default value if evaluator config is not found
        :return: evaluator config
        """
        try:
            return api.is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config,
                                                                       cls.get_name(),
                                                                       raise_errors=True)
        except KeyError:
            for parent in cls.mro():
                try:
                    return api.is_tentacle_activated_in_tentacles_setup_config(tentacles_setup_config,
                                                                               parent.__name__,
                                                                               raise_errors=True)
                except KeyError:
                    pass
        return default

    def save_evaluation_expiration_time(self, eval_note_time_to_live, eval_note_changed_time=None) -> None:
        """
        Use only if the current evaluation is to stay for a pre-defined amount of seconds
        :param eval_note_time_to_live:
        :param eval_note_changed_time:
        :return: None
        """
        self.eval_note_time_to_live = eval_note_time_to_live
        self.eval_note_changed_time = eval_note_changed_time if eval_note_changed_time else time.time()

    def eval_note_changed(self) -> None:
        """
        Eval note changed callback
        :return: None
        """
        if self.eval_note_time_to_live is not None and self.eval_note_changed_time is None:
            self.eval_note_changed_time = time.time()

    def ensure_eval_note_is_not_expired(self) -> None:
        """
        Eval note expiration check
        :return: None
        """
        if self.eval_note_time_to_live is not None:
            if self.eval_note_changed_time is None:
                self.eval_note_changed_time = time.time()

            if time.time() - self.eval_note_changed_time > self.eval_note_time_to_live:
                self.eval_note = common_constants.START_PENDING_EVAL_NOTE
                self.eval_note_time_to_live = None
                self.eval_note_changed_time = None

    def get_exchange_symbol_data(self, exchange_name: str, exchange_id: str, symbol: str):
        try:
            import octobot_trading.api as exchange_api
            exchange_manager = exchange_api.get_exchange_manager_from_exchange_name_and_id(exchange_name, exchange_id)
            return exchange_api.get_symbol_data(exchange_manager, symbol)
        except (ImportError, KeyError):
            self.logger.error(f"Can't get {exchange_name} from exchanges instances")
        return

    async def evaluators_callback(self,
                                  matrix_id,
                                  evaluator_name,
                                  evaluator_type,
                                  exchange_name,
                                  cryptocurrency,
                                  symbol,
                                  time_frame,
                                  data):
        # Used to communicate between evaluators
        pass
